#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import platform
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    test = test_gbp_l3p_func()
    if test.test_gbp_l3p_func_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_FUNC_1')
    if test.test_gbp_l3p_func_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_FUNC_2')
    if test.test_gbp_l3p_func_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_FUNC_3')
    if test.test_gbp_l3p_func_4() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_L3P_FUNC_4')
    test.cleanup()
    utils_libs.report_results('test_gbp_l3p_func', 'test_results.txt')
    sys.exit(1)


class test_gbp_l3p_func(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_l3p_func.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_l3p_func.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info("\n## START OF GBP L3_POLICY FUNCTIONALITY TESTSUITE\n")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.l3p_name = 'demo_l3p'
        self.l2p_name = 'demo_l2p'

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('## %s: FAILED' % (tc_name))
        for obj in ['group', 'l2p', 'l3p']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_l3p_func_1(
            self,
            name_uuid='',
            l3p_uuid='',
            rep_cr=0,
            rep_del=0):

        if rep_cr == 0 and rep_del == 0:
            self._log.info(
                "\n########################################################\n"
                "TESTCASE_GBP_L3P_FUNC_1: TO CREATE/VERIFY/DELETE/VERIFY a "
                "L3POLICY with DEFAULT ATTRIB VALUE\n"
                "TEST_STEPS::\n"
                "Create L3 Policy Object\n"
                "Verify the attributes & value, show & list cmds\n"
                "Verify the implicit neutron objects\n"
                "Delete L3 Policy Object\n"
                "Verify that PR and implicit neutron objects has got "
                "deleted, show & list cmds\n"
                "##########################################################\n")

        if name_uuid == '':
            name_uuid = self.l3p_name
        # Testcase work-flow starts
        if rep_cr == 0 or rep_cr == 1:
            self._log.info(
                '\n## Step 1: Create L3Policy with default attrib vals##\n')
            l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(1, 'l3p', name_uuid)
            if l3p_uuid == 0:
                self._log.info("\n## Step 1: Create L3Policy == Failed")
                return 0
            # default subnet= 10.0.0.0/8 & subnet_prefix_length= 24
            self._log.info('# Step 2A: Verify L3Policy using -list cmd')
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    0, 'l3p', l3p_uuid, name_uuid, '10.0.0.0/8', '24') == 0:
                self._log.info(
                    "\n## Step 2A: Verify L3Policy using -list option "
                    "== Failed")
                return 0
            self._log.info('# Step 2B: Verify L3Policy using -show cmd')
            if 'Ubuntu' in platform.linux_distribution():  # Only for devstack
                rtr_uuid = self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1,
                    'l3p',
                    l3p_uuid,
                    ret='default',
                    id=l3p_uuid,
                    name=name_uuid,
                    ip_pool='10.0.0.0/8',
                    subnet_prefix_length='24',
                    ip_version='4')
                if rtr_uuid != 0 and isinstance(rtr_uuid, str):
                    rtr_name = 'l3p_%s' % (name_uuid)
                    if self.gbpverify.neut_ver_all(
                            'router',
                            rtr_uuid,
                            name=rtr_name,
                            admin_state_up='True',
                            status='ACTIVE') == 0:
                        self._log.info(
                            "\n## Step 2D: Verify L3Policy using -show "
                            "option == Failed")
                        return 0
                else:
                    self._log.info(
                        "\n## Step 2C: Verify L3Policy using -show "
                        "option == Failed")
                    return 0
            else:
                if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                        1,
                        'l3p',
                        l3p_uuid,
                        id=l3p_uuid,
                        name=name_uuid,
                        ip_pool='10.0.0.0/8',
                        subnet_prefix_length='24',
                        ip_version='4') == 0:
                    self._log.info(
                        "\n## Step 2C: Verify L3Policy using -show "
                        "option == Failed")
                    return 0
        #######
        if rep_del == 0 or rep_del == 1:
            self._log.info('\n## Step 3: Delete L3Policy using name  ##\n')
            if self.gbpcfg.gbp_policy_cfg_all(0, 'l3p', name_uuid) == 0:
                self._log.info("\n## Step 3: Delete L3Policy == Failed")
                return 0
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    0, 'l3p', name_uuid, l3p_uuid) != 0:
                self._log.info(
                    "\n## Step 3A: Verify L3Policy is Deleted using "
                    "-list option == Failed")
                return 0
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'l3p', name_uuid, l3p_uuid) != 0:
                self._log.info(
                    "\n## Step 3B: Verify L3Policy is Deleted using "
                    "-show option == Failed")
                return 0
            if rep_cr == 0 and rep_del == 0:
                self._log.info("\n## TESTCASE_GBP_L3P_FUNC_1: PASSED")
        return 1

    def test_gbp_l3p_func_2(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_L3P_FUNC_2: TO CREATE/UPDATE/DELETE/VERIFY a "
            "L3POLICY with EDITABLE ATTRs\n"
            "TEST_STEPS::\n"
            "Create L3Policy Object with non-default params\n"
            "Verify the attributes & value, show & list cmds\n"
            "Update the L3Policy Objects\n"
            "Verify the attributes & value, show & list cmds\n"
            "Delete L3Policy using Name\n"
            "Verify that L3P has got deleted, show & list cmds\n"
            "##############################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Policy L3Policy with non-default "
            "attrs and values ##")
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'l3p', self.l3p_name, ip_pool='20.20.0.0/24',
            subnet_prefix_length='28')
        if l3p_uuid == 0:
            self._log.info("\n## Step 1: Create L3Policy == Failed")
            return 0
        self._log.info('\n## Step 2B: Verify L3Policy using -show cmd')
        if 'Ubuntu' in platform.linux_distribution():  # Only for devstack
            rtr_uuid = self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                l3p_uuid,
                ret='default',
                id=l3p_uuid,
                name=self.l3p_name,
                ip_pool='20.20.0.0/24',
                subnet_prefix_length='28',
                ip_version='4')
            if rtr_uuid != 0 and isinstance(rtr_uuid, str):
                rtr_name = 'l3p_%s' % (self.l3p_name)
                if self.gbpverify.neut_ver_all(
                        'router',
                        rtr_uuid,
                        name=rtr_name,
                        admin_state_up='True',
                        status='ACTIVE') == 0:
                    self._log.info(
                        "\n## Step 2D: Verify L3Policy using -show option"
                        " == Failed")
                    return 0
            else:
                self._log.info(
                    "\n## Step 2C: Verify L3Policy using -show option"
                    " == Failed")
                return 0
        else:
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1,
                    'l3p',
                    l3p_uuid,
                    id=l3p_uuid,
                    name=self.l3p_name,
                    ip_pool='20.20.0.0/24',
                    subnet_prefix_length='28',
                    ip_version='4') == 0:
                self._log.info(
                    "\n## Step 2C: Verify L3Policy using -show option"
                    " == Failed")

        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'l3p', self.l3p_name, subnet_prefix_length='26') == 0:
            self._log.info(
                "\n## Step 3: UPdating L3Policy attributes == Failed")
            return 0
        self._log.info(
            "\n## Step 3: Verify that Updated Attributes in L3Policy")
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                self.l3p_name,
                id=l3p_uuid,
                name=self.l3p_name,
                ip_pool='20.20.0.0/24',
                subnet_prefix_length='26',
                ip_version='4') == 0:
            self._log.info(
                "\n## Step 3: Verify L3Policy using -show option == Failed")
        self.test_gbp_l3p_func_1(name_uuid=l3p_uuid, rep_cr=2)
        self._log.info("\n## TESTCASE_GBP_L3P_FUNC_2: PASSED")
        return 1

    def test_gbp_l3p_func_3(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_L3P_FUNC_3: TO CREATE/UPDATE/DELETE/VERIFY "
            "L3POLICY AND ASSOCIATED L2POLICY\n"
            "TEST_STEPS::\n"
            "Create L3Policy with defined attributes\n"
            "Create L2Policy with default attributes\n"
            "Update L2Policy to change the from default to the above "
            "non-default L3Policy\n"
            "Verify the Update of L3Policy attribute of L2Policy fails\n"
            "Update L3Policy(default) editable attributes\n"
            "Delete the L2Policy(this causes auto-delete of default-L3Pol)\n"
            "Verify L3/L2Policies successfully deleted\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create L2 L3 Policy
        self._log.info(
            "\n## Step 1: Create L3Policy with non-default attrs and "
            "values ##")
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'l3p', self.l3p_name, ip_pool='20.20.0.0/24',
            subnet_prefix_length='28',
            proxy_ip_pool='192.167.0.0/16')
        if l3p_uuid == 0:
            self._log.info("\n## Step 1: Create L3Policy == Failed")
            return 0
        self._log.info(
            '\n## Step 1A: Create L2Policy with default attributes##\n')
        l2p = self.gbpcfg.gbp_policy_cfg_all(1, 'l2p', self.l2p_name)
        if l2p == 0:
            self._log.info(
                "\n## New L2Policy Create Failed, hence "
                "Testcase_gbp_l3p_func_3 ABORTED\n")
            return 0
        elif len(l2p) < 2:
            self._log.info(
                "\n## New L2Policy Create Failed due to "
                "unexpected tuple length\n")
            return 0
        else:
            l2p_uuid, def_l3p_uuid = l2p[0], l2p[1]
        # Associating L2Policy with non-default L3Policy(should Fail) and
        # UPdating the L3Policy(in-use/default)
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'l2p', self.l2p_name, l3_policy_id=l3p_uuid) != 0:
            self._log.info(
                "\n## Updating/Changing L3Policy attribute of "
                "L2Policy did NOT Fail")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1, 'l2p', self.l2p_name, l3_policy_id=def_l3p_uuid) == 0:
            self._log.info(
                "\n## Step 3A: Verify L2Policy is still associated to "
                "its  default L3Policy == Failed")
            return 0
        if self.gbpcfg.gbp_policy_cfg_all(
                2, 'l3p', def_l3p_uuid, subnet_prefix_length='27') == 0:
            self._log.info(
                "\n## Step 4: UPdating default L3Policy's "
                "attributes == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                def_l3p_uuid,
                id=def_l3p_uuid,
                ip_pool='10.0.0.0/8',
                l2_policies=l2p_uuid,
                subnet_prefix_length='27',
                ip_version='4') == 0:
            self._log.info(
                "\n## Step 4A: Verify L3Policy after associating "
                "to the L2Policy == Failed")
            return 0

        # Delete L2/L3 Policies
        if self.gbpcfg.gbp_policy_cfg_all(0, 'l2p', l2p_uuid) == 0:
            self._log.info("\n## Step 5: Delete L2Policy == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l2p', l2p_uuid) != 0:
            self._log.info("\n## Step 5A: Verify Delete of L2Policy == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l3p', def_l3p_uuid) != 0:
            self._log.info(
                "\n## Step 5B: Verify Auto-Delete of default "
                "L3Policy == Failed")
            return 0
        self._log.info("\n## TESTCASE_GBP_L3P_FUNC_3: PASSED")
        return 1

    def test_gbp_l3p_func_4(self):

        self._log.info(
            "\n############################################################\n"
            "TESTCASE_GBP_L3P_FUNC_4: TO CREATE/UPDATE/DELETE/VERIFY "
            "MULTI L2POLICY to SINGLE L3POLICY\n"
            "TEST_STEPS::\n"
            "Create non-default L3Policy with defined attributes\n"
            "Create Multiple L2Policies with above non-default L3policy\n"
            "Verify that L2Policies are created with non-default L3Policy\n"
            "Delete all L2 Policies\n"
            "Verify that non-default L3 Policy exists but with null "
            "L2Policies\n"
            "Delete the L3Policy\n"
            "Verify L3/L2Policys successfully deleted\n"
            "##############################################################\n")

        # Testcase work-flow starts
        # Create and Verify non-default L3 Policy
        self._log.info(
            "\n## Step 1: Create Policy L3Policy with non-default "
            "attrs and values ")
        l3p_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'l3p', self.l3p_name, ip_pool='40.50.0.0/16',
            subnet_prefix_length='25')
        if l3p_uuid == 0:
            self._log.info("\n## Step 1: Create L3Policy == Failed")
            return 0
        if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                1,
                'l3p',
                l3p_uuid,
                id=l3p_uuid,
                name=self.l3p_name,
                ip_pool='40.50.0.0/16',
                subnet_prefix_length='25',
                ip_version='4') == 0:
            self._log.info("\n## Step 1A: Verify non-default == Failed")
            return 0
        # Create and verify multiple L2 policy with above non-default L3P
        self._log.info(
            "\n## Step 2: Create and Verify multiple(n=10) L2Policy "
            "associated with 1 non-default L3P")
        l2p_uuid_list = []
        n, i = 11, 1
        while i < n:
            l2p_name = 'demo_l2p_%s' % (i)
            l2p = self.gbpcfg.gbp_policy_cfg_all(
                1, 'l2p', l2p_name, l3_policy_id=l3p_uuid)
            if l2p == 0:
                self._log.info(
                    "\n## Step 2B:New L2Policy Create Failed, hence "
                    "Testcase_gbp_l3p_func_4 ABORTED\n")
                return 0
            elif len(l2p) < 2:
                self._log.info(
                    "\n## Step 2C: New L2Policy Create Failed due to "
                    "unexpected tuple length\n")
                return 0
            else:
                l2p_uuid = l2p[0]
                l2p_uuid_list.append(l2p_uuid)
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(
                    1, 'l2p', l2p_name, id=l2p_uuid,
                    l3_policy_id=l3p_uuid) == 0:
                self._log.info(
                    "\n## Step 2D: Verify L2Policy using non-default "
                    "L3P == Failed")
                return 0
            i += 1
        # Verify that non-default L3P has all the above create L2Ps
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'l3p', l3p_uuid, 'l2_policies', l2p_uuid_list) == 0:
            self._log.info(
                "\n## Step 2E: Verifying multiple L2Ps mapped to "
                "non-default L3P == Failed \n")
            return 0
        # Delete all L2Ps and verify that non-default L3P has null L2Ps
        self._log.info(
            "\n## Step 3: Delete all L2Ps and verify that non-default "
            "L3P has no L2P associated\n")
        for l2pid in l2p_uuid_list:
            if self.gbpcfg.gbp_policy_cfg_all(0, 'l2p', l2pid) == 0:
                self._log.info(
                    "\n## Step 3: Delete of L2P %s == Failed\n" %
                    (l2pid))
                return 0
            if self.gbpverify.gbp_l2l3ntk_pol_ver_all(1, 'l2p', l2pid) != 0:
                self._log.info(
                    "\n## Step 3A: Verify that L2P got deleted == Failed\n")
                return 0
        if self.gbpverify.gbp_obj_ver_attr_all_values(
                'l3p', l3p_uuid, 'l2_policies', l2p_uuid_list) != 0:
            self._log.info(
                "\n## Step 3B: Verifying Non-Default L3P has no more "
                "L2P mapped == Failed \n")
            return 0
        self.test_gbp_l3p_func_1(name_uuid=l3p_uuid, rep_cr=2)
        self._log.info("\n## TESTCASE_GBP_L3P_FUNC_4: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import oslo_messaging as messaging
import six

from gbpservice.contrib.nfp.configurator.agents import agent_base
from gbpservice.contrib.nfp.configurator.lib import (
    nfp_service_constants as const)
from gbpservice.contrib.nfp.configurator.lib import utils as load_driver
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)


class ConfigScriptRpcManager(agent_base.AgentBaseRPCManager):
    """ Implements ConfigScriptRpcManager class which receives requests
        from Configurator module.

    Methods of this class are invoked by the configurator. Events are
    created according to the requests received and enqueued to worker queues.

    """

    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, sc, conf):
        """Instantiates child and parent class objects.

        :param sc: Service Controller object that is used to communicate
        with process model core file.
        :param conf: Configuration object that is used for configuration
        parameter access.

        """

        super(ConfigScriptRpcManager, self).__init__(sc, conf)

    def run_nfp_service(self, context, resource_data):
        """ Receives request to execute config script.

        :param context: RPC context
        :param kwargs: Contains configuration script and request information

        """

        msg = ("ConfigScriptRpcManager received Create Heat request.")
        LOG.debug(msg)

        arg_dict = {'context': context,
                    'resource_data': resource_data}
        ev = self.sc.new_event(id=const.CREATE_NFP_SERVICE_EVENT,
                               data=arg_dict, key=None)
        self.sc.post_event(ev)


class ConfigScriptEventHandler(agent_base.AgentBaseEventHandler):
    """ Handler class which invokes nfp_service driver methods

    Worker processes dequeue the worker queues and invokes the
    appropriate handler class methods for ConfigScript methods.

    """

    def __init__(self, sc, drivers, rpcmgr):
        """ Initializes parent and child class objects.

        :param sc: Service Controller object that is used to communicate
        with process model.
        :param drivers: Dictionary of driver name to object mapping
        :param rpcmgr: ConfigScriptRpcManager class object

        """

        super(ConfigScriptEventHandler, self).__init__(sc, drivers, rpcmgr)
        self.sc = sc
        self.drivers = drivers
        self.rpcmgr = rpcmgr

    def _get_driver(self):
        """ Retrieves driver object given the service type.

        """

        driver_id = const.SERVICE_TYPE
        return self.drivers[driver_id]

    def handle_event(self, ev):
        """ Demultiplexes the nfp_service request to appropriate
        driver methods.

        :param ev: Event object sent from process model event handler

        """

        try:
            agent_info = ev.data['context']
            notification_context = agent_info['context']
            resource = agent_info['resource']
            resource_data = ev.data['resource_data']

            msg = ("Worker process with ID: %s starting to "
                   "handle task: %s of type ConfigScript. "
                   % (os.getpid(), ev.id))
            LOG.debug(msg)

            driver = self._get_driver()
            self.method = getattr(driver, "run_%s" % resource)

            result = self.method(notification_context, resource_data)
        except Exception as err:
            result = const.ERROR_RESULT
            msg = ("Failed to handle event: %s. %s"
                   % (ev.id, str(err).capitalize()))
            LOG.error(msg)
        finally:
            del agent_info['notification_data']
            del agent_info['service_vendor']
            service_type = agent_info.pop('resource_type')

            if result in const.UNHANDLED_RESULT:
                data = {'status_code': const.UNHANDLED_RESULT}
            else:
                data = {'status_code': const.FAILURE,
                        'error_msg': result}

            msg = {'info': {'service_type': service_type,
                            'context': notification_context},
                   'notification': [{'resource': resource,
                                     'data': data}]
                   }

            self.notify._notification(msg)


def events_init(sc, drivers, rpcmgr):
    """Registers events with core service controller.

    All the events will come to handle_event method of class instance
    registered in 'handler' field.

    :param drivers: Driver instances registered with the service agent
    :param rpcmgr: Instance to receive all the RPC messages from configurator
    module.

    Returns: None

    """

    event = nfp_event.Event(
        id=const.CREATE_NFP_SERVICE_EVENT,
        handler=ConfigScriptEventHandler(sc, drivers, rpcmgr))
    sc.register_events([event])


def load_drivers(conf):
    """Imports all the driver files corresponding to this agent.

    Returns: Dictionary of driver objects with a specified service type and
    vendor name

    """

    ld = load_driver.ConfiguratorUtils(conf)
    drivers = ld.load_drivers(const.SERVICE_TYPE)

    for service_type, driver_name in six.iteritems(drivers):
        driver_obj = driver_name(conf=conf)
        drivers[service_type] = driver_obj

    return drivers


def register_service_agent(cm, sc, conf, rpcmgr):
    """Registers ConfigScript service agent with configurator module.

    :param cm: Instance of configurator module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration
    :param rpcmgr: Instance containing RPC methods which are invoked by
    configurator module on corresponding RPC message arrival

    """

    service_type = const.SERVICE_TYPE
    cm.register_service_agent(service_type, rpcmgr)


def init_agent(cm, sc, conf):
    """Initializes Config Script agent.

    :param cm: Instance of configuration module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration

    """

    try:
        drivers = load_drivers(conf)
    except Exception as err:
        msg = ("Config Script failed to load drivers. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Config Script loaded drivers successfully.")
        LOG.debug(msg)

    rpcmgr = ConfigScriptRpcManager(sc, conf)
    try:
        events_init(sc, drivers, rpcmgr)
    except Exception as err:
        msg = ("Config Script Events initialization unsuccessful. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Config Script Events initialization successful.")
        LOG.debug(msg)

    try:
        register_service_agent(cm, sc, conf, rpcmgr)
    except Exception as err:
        msg = ("Config Script service agent registration unsuccessful. %s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise Exception(err)
    else:
        msg = ("Config Script service agent registration successful.")
        LOG.debug(msg)

    msg = ("ConfigScript as a Service Module Initialized.")
    LOG.info(msg)


def init_agent_complete(cm, sc, conf):
    """ Initializes periodic tasks

    """

    msg = (" Config Script agent init complete")
    LOG.info(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""nfp_network_function_device_interface

Revision ID: 9cedbcd3e9ee
Revises: 5629167be1d1
Create Date: 2016-10-24 19:52:17.140960

"""

# revision identifiers, used by Alembic.
revision = '9cedbcd3e9ee'
down_revision = '5629167be1d1'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'nfp_network_function_device_interfaces',
        sa.Column('tenant_id', sa.String(length=255), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('plugged_in_port_id', sa.String(length=36), nullable=True),
        sa.Column('interface_position',
                  sa.Integer(),
                  nullable=True),
        sa.Column('mapped_real_port_id', sa.String(length=36), nullable=True),
        sa.Column('network_function_device_id',
                  sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['plugged_in_port_id'],
                                ['nfp_port_infos.id'],
                                ondelete='SET NULL'),
        sa.ForeignKeyConstraint(['network_function_device_id'],
                                ['nfp_network_function_devices.id'],
                                ondelete='SET NULL'),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import exceptions

from neutron_lib import context as n_context
from neutron_lib import rpc as n_rpc
from oslo_config import cfg
from oslo_config import cfg as oslo_config
import oslo_messaging as messaging
from oslo_serialization import jsonutils
import requests
import six

from gbpservice._i18n import _
from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.lib import rest_client_over_unix as unix_rc

LOG = nfp_logging.getLogger(__name__)
Version = 'v1'  # v1/v2/v3#

rest_opts = [
    cfg.StrOpt('rest_server_address',
               default='', help='Rest connection IpAddr'),
    cfg.IntOpt('rest_server_port',
               default=8080, help='Rest connection Port'),
]

rpc_opts = [
    cfg.StrOpt('topic',
               default='', help='Topic for rpc connection'),
]

oslo_config.CONF.register_opts(rest_opts, "REST")
oslo_config.CONF.register_opts(rpc_opts, "RPC")
n_rpc.init(cfg.CONF)

UNIX_REST = 'unix_rest'
TCP_REST = 'tcp_rest'

""" Common Class for restClient exceptions """


class RestClientException(exceptions.Exception):

    """ RestClient Exception """


""" Common Class to handle restclient request"""


class RestApi(object):

    def __init__(self, rest_server_address, rest_server_port):
        self.rest_server_address = rest_server_address
        self.rest_server_port = rest_server_port
        self.url = "http://%s:%s/v1/nfp/%s"

    def _response(self, resp, url):
        success_code = [200, 201, 202, 204]
        # Evaluate responses into success and failures.
        # Raise exception for failure cases which needs
        # to be handled in caller function.
        if success_code.__contains__(resp.status_code):
            return resp
        elif resp.status_code == 400:
            raise RestClientException(_("HTTPBadRequest: %s") % resp.reason)
        elif resp.status_code == 401:
            raise RestClientException(_("HTTPUnauthorized: %s") % resp.reason)
        elif resp.status_code == 403:
            raise RestClientException(_("HTTPForbidden: %s") % resp.reason)
        elif resp.status_code == 404:
            raise RestClientException(_("HttpNotFound: %s") % resp.reason)
        elif resp.status_code == 405:
            raise RestClientException(
                _("HTTPMethodNotAllowed: %s") % resp.reason)
        elif resp.status_code == 406:
            raise RestClientException(_("HTTPNotAcceptable: %s") % resp.reason)
        elif resp.status_code == 408:
            raise RestClientException(
                _("HTTPRequestTimeout: %s") % resp.reason)
        elif resp.status_code == 409:
            raise RestClientException(_("HTTPConflict: %s") % resp.reason)
        elif resp.status_code == 415:
            raise RestClientException(
                _("HTTPUnsupportedMediaType: %s") % resp.reason)
        elif resp.status_code == 417:
            raise RestClientException(
                _("HTTPExpectationFailed: %s") % resp.reason)
        elif resp.status_code == 500:
            raise RestClientException(_("HTTPServerError: %s") % resp.reason)
        else:
            raise RestClientException(_('Unhandled Exception code: '
                                        '%(status)s %(reason)s') %
                                      {'status': resp.status_code,
                                       'reason': resp.reason})
        return resp

    def post(self, path, body, method_type):
        """Post restclient request handler
        Return:Http response
        """
        url = self.url % (
            self.rest_server_address,
            self.rest_server_port, path)
        data = jsonutils.dumps(body)
        try:
            # Method-Type needs to be added here,as DELETE/CREATE
            # both case are handled by post as delete also needs
            # to send data to the rest-server.
            headers = {"content-type": "application/json",
                       "method-type": method_type}
            resp = requests.post(url, data,
                                 headers=headers)
            message = "POST url %s %d" % (url, resp.status_code)
            LOG.info(message)
            return self._response(resp, url)
        except RestClientException as rce:
            message = "Rest API %s - Failed. Reason: %s" % (
                url, rce)
            LOG.error(message)

    def put(self, path, body):
        """Put restclient request handler
        Return:Http response
        """
        url = self.url % (
            self.rest_server_address,
            self.rest_server_port, path)
        data = jsonutils.dumps(body)
        try:
            headers = {"content-type": "application/json"}
            resp = requests.put(url, data,
                                headers=headers)
            message = "PUT url %s %d" % (url, resp.status_code)
            LOG.info(message)
            return self._response(resp, url)
        except RestClientException as rce:
            message = "Rest API %s - Failed. Reason: %s" % (
                url, rce)
            LOG.error(message)

    def get(self, path):
        """Get restclient request handler
        Return:Http response
        """
        url = self.url % (
            self.rest_server_address,
            self.rest_server_port, path)
        try:
            headers = {"content-type": "application/json"}
            resp = requests.get(url,
                                headers=headers)
            message = "GET url %s %d" % (url, resp.status_code)
            LOG.info(message)
            return self._response(resp, url)
        except RestClientException as rce:
            message = "Rest API %s - Failed. Reason: %s" % (
                url, rce)
            LOG.error(message)


""" Common Class to handle rpcclient request"""


class RPCClient(object):
    API_VERSION = '1.0'

    def __init__(self, topic):
        self.topic = topic
        target = messaging.Target(topic=self.topic,
                                  version=self.API_VERSION)
        self.client = n_rpc.get_client(target)
        self.cctxt = self.client.prepare(version=self.API_VERSION,
                                         topic=self.topic)


def send_request_to_configurator(conf, context, body,
                                 method_type, device_config=False,
                                 network_function_event=False):
    """Common function to handle (create, delete) request for configurator.
    Send create/delete to configurator rest-server.
    Return:Http Response
    """
    # This function reads configuration data and decides
    # method (tcp_rest/rpc) for sending request to configurator.
    if device_config:
        method_name = method_type.lower() + '_network_function_device_config'
        body['info']['context'].update({'neutron_context': context.to_dict()})
    elif network_function_event:
        method_name = 'network_function_event'
    else:
        if (body['config'][0]['resource'] in
                nfp_constants.CONFIG_TAG_RESOURCE_MAP.values()):
            body['config'][0]['resource_data'].update(
                {'neutron_context': context.to_dict()})
            body['info']['context'].update(
                {'neutron_context': context.to_dict()})
        method_name = method_type.lower() + '_network_function_config'

    if conf.backend == TCP_REST:
        try:
            rc = RestApi(conf.REST.rest_server_address,
                         conf.REST.rest_server_port)
            if method_type.lower() in [nfp_constants.CREATE,
                                       nfp_constants.DELETE]:
                resp = rc.post(method_name, body, method_type.upper())
                message = "%s -> POST response: (%s) body: %s " % (method_name,
                                                                   resp, body)
                LOG.debug(message)
            elif method_type.lower() in [nfp_constants.UPDATE]:
                resp = rc.put(method_name, body)
                message = "%s -> PUT response: (%s) body: %s " % (method_name,
                                                                  resp, body)
                LOG.debug(message)
            else:
                message = ("%s api not supported" % (method_name))
                LOG.error(message)
        except RestClientException as rce:
            message = "%s -> POST request failed.Reason: %s" % (
                method_name, rce)
            LOG.error(message)

    elif conf.backend == UNIX_REST:
        try:
            if method_type.lower() in [nfp_constants.CREATE,
                                       nfp_constants.DELETE]:
                resp, content = unix_rc.post(method_name,
                                             body=body)
                message = ("%s -> POST response: (%s) body : %s " %
                           (method_name, content, body))
                LOG.debug(message)
            elif method_type.lower() in [nfp_constants.UPDATE]:
                resp, content = unix_rc.put(method_name,
                                            body=body)
                message = ("%s -> PUT response: (%s) body : %s " %
                           (method_name, content, body))
                LOG.debug(message)
            else:
                message = ("%s api not supported" % (method_name))
                LOG.error(message)
        except unix_rc.RestClientException as rce:
            message = "%s -> request failed . Reason %s " % (
                method_name, rce)
            LOG.error(message)

    else:
        message = ("%s -> RPC request sent. " % (method_name))
        LOG.info(message)
        rpcClient = RPCClient(conf.RPC.topic)
        rpcClient.cctxt.cast(context, method_name,
                             body=body)


def get_response_from_configurator(conf):
    """Common function to handle get request for configurator.
    Get notification http response from configurator rest server.
    Return:Http Response
    response_data = [
            {'receiver': <neutron/device_orchestrator/service_orchestrator>,
             'resource': <firewall/vpn/loadbalancer/orchestrator>,
             'method': <notification method name>,
             'kwargs': <notification method arguments>
        },
    ]
    """
    # This function reads configuration data and decides
    # method (tcp_rest/ unix_rest/ rpc) for get response from configurator.
    if conf.backend == TCP_REST:
        try:
            rc = RestApi(conf.REST.rest_server_address,
                         conf.REST.rest_server_port)
            resp = rc.get('get_notifications')
            rpc_cbs_data = jsonutils.loads(resp.content)
            return rpc_cbs_data
        except RestClientException as rce:
            message = ("get_notification ->"
                       "GET request failed. Reason : %s" % (rce))
            LOG.error(message)
            return "get_notification -> GET request failed. Reason : %s" % (
                rce)
        except Exception as e:
            message = ("get_notification ->"
                       "GET request failed. Reason : %s" % (e))
            LOG.error(message)
            return "get_notification -> GET request failed. Reason : %s" % (
                e)

    elif conf.backend == UNIX_REST:
        try:
            resp, content = unix_rc.get('get_notifications')
            content = jsonutils.loads(content)
            if content:
                message = ("get_notification ->"
                           "GET response: (%s)" % (content))
                LOG.debug(message)
            return content
        except unix_rc.RestClientException as rce:
            message = ("get_notification ->"
                       "GET request failed. Reason : %s" % (
                           rce))
            LOG.error(message)
            return "get_notification -> GET request failed. Reason : %s" % (
                rce)
        except Exception as e:
            message = ("get_notification ->"
                       "GET request failed. Reason : %s" % (
                           e))
            LOG.error(message)
            return "get_notification -> GET request failed. Reason : %s" % (
                e)

    else:
        rpc_cbs_data = []
        try:
            rpcClient = RPCClient(conf.RPC.topic)
            context = n_context.Context(
                'config_agent_user', 'config_agent_tenant')
            rpc_cbs_data = rpcClient.cctxt.call(context,
                                                'get_notifications')
            return rpc_cbs_data
        except Exception as e:
            message = "Exception while processing %s" % e
            LOG.error(message)
            return "get_notification -> GET request failed. Reason : %s" % (
                e)


def parse_service_flavor_string(service_flavor_str):
    """Parse service_flavour string to service details dictionary.
        Return: Service Details Dictionary
    """
    service_details = {}
    if ',' not in service_flavor_str:
        service_details['device_type'] = 'nova'
        service_details['service_vendor'] = service_flavor_str
    else:
        service_flavor_dict = dict(item.split('=') for item
                                   in service_flavor_str.split(','))
        service_details = {key.strip(): value.strip() for key, value
                           in six.iteritems(service_flavor_dict)}
    return service_details



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_ctx
from oslo_config import cfg
from sqlalchemy.orm import exc as orm_exc

from gbpservice.neutron.db.grouppolicy.extensions import group_proxy_db
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_extension_driver_api as test_ext_base)


class ExtensionDriverTestCaseMixin(object):

    def test_proxy_group_extension(self):
        l3p = self.create_l3_policy()['l3_policy']
        self.assertEqual('192.168.0.0/16', l3p['proxy_ip_pool'])
        self.assertEqual(28, l3p['proxy_subnet_prefix_length'])

        l2p = self.create_l2_policy(l3_policy_id=l3p['id'])['l2_policy']
        ptg = self.create_policy_target_group(
            l2_policy_id=l2p['id'])['policy_target_group']
        self.assertIsNone(ptg['proxy_group_id'])
        self.assertIsNone(ptg['proxied_group_id'])
        self.assertIsNone(ptg['proxy_type'])

        # Verify Default L3P pool mapping on show
        l3p = self.show_l3_policy(l3p['id'])['l3_policy']
        self.assertEqual('192.168.0.0/16', l3p['proxy_ip_pool'])
        self.assertEqual(28, l3p['proxy_subnet_prefix_length'])

        # No proxy
        ptg_proxy = self.create_policy_target_group()['policy_target_group']
        self.assertTrue(ptg_proxy['enforce_service_chains'])
        ptg_proxy = self.show_policy_target_group(
            ptg_proxy['id'])['policy_target_group']
        self.assertTrue(ptg_proxy['enforce_service_chains'])
        ptg_proxy = self.create_policy_target_group(
            enforce_service_chains=False,
            is_admin_context=True)['policy_target_group']
        self.assertFalse(ptg_proxy['enforce_service_chains'])
        ptg_proxy = self.show_policy_target_group(
            ptg_proxy['id'])['policy_target_group']
        self.assertFalse(ptg_proxy['enforce_service_chains'])

        ptg_proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'])['policy_target_group']
        self.assertIsNone(ptg_proxy['proxy_group_id'])
        self.assertEqual(ptg['id'], ptg_proxy['proxied_group_id'])
        self.assertEqual('l3', ptg_proxy['proxy_type'])
        self.assertFalse(ptg_proxy['enforce_service_chains'])

        # Verify relationship added
        ptg = self.show_policy_target_group(ptg['id'])['policy_target_group']
        self.assertEqual(ptg_proxy['id'], ptg['proxy_group_id'])
        self.assertIsNone(ptg['proxied_group_id'])

        pt = self.create_policy_target(
            policy_target_group_id=ptg_proxy['id'])['policy_target']
        self.assertFalse(pt['proxy_gateway'])
        self.assertFalse(pt['group_default_gateway'])
        pt = self.create_policy_target(
            policy_target_group_id=ptg_proxy['id'],
            proxy_gateway=True, group_default_gateway=True)['policy_target']
        self.assertTrue(pt['proxy_gateway'])
        self.assertTrue(pt['group_default_gateway'])
        pt = self.show_policy_target(pt['id'])['policy_target']
        self.assertTrue(pt['proxy_gateway'])
        self.assertTrue(pt['group_default_gateway'])

    def test_preexisting_pt(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        pt = self.create_policy_target(
            policy_target_group_id=ptg['id'])['policy_target']
        self.assertTrue('proxy_gateway' in pt)
        self.assertTrue('group_default_gateway' in pt)

        # Forcefully delete the entry in the proxy table, and verify that it's
        # fixed by the subsequent GET
        admin_context = n_ctx.get_admin_context()
        mapping = admin_context.session.query(
            group_proxy_db.ProxyGatewayMapping).filter_by(
                policy_target_id=pt['id']).one()
        admin_context.session.delete(mapping)
        query = admin_context.session.query(
            group_proxy_db.ProxyGatewayMapping).filter_by(
                policy_target_id=pt['id'])
        self.assertRaises(orm_exc.NoResultFound, query.one)

        # Showing the object just ignores the extension
        pt = self.show_policy_target(pt['id'],
                                     expected_res_status=200)['policy_target']
        self.assertFalse('proxy_gateway' in pt)
        self.assertFalse('group_default_gateway' in pt)

        # Updating the object just ignores the extension
        pt = self.update_policy_target(
            pt['id'], name='somenewname',
            expected_res_status=200)['policy_target']
        self.assertEqual('somenewname', pt['name'])
        self.assertFalse('proxy_gateway' in pt)
        self.assertFalse('group_default_gateway' in pt)

    def test_proxy_group_multiple_proxies(self):
        # same PTG proxied multiple times will fail
        ptg = self.create_policy_target_group()['policy_target_group']
        self.create_policy_target_group(proxied_group_id=ptg['id'])
        # Second proxy will fail
        res = self.create_policy_target_group(proxied_group_id=ptg['id'],
                                              expected_res_status=400)
        self.assertEqual('InvalidProxiedGroup', res['NeutronError']['type'])

    def test_proxy_group_chain_proxy(self):
        # Verify no error is raised when chaining multiple proxy PTGs
        ptg0 = self.create_policy_target_group()['policy_target_group']
        ptg1 = self.create_policy_target_group(
            proxied_group_id=ptg0['id'],
            expected_res_status=201)['policy_target_group']
        self.create_policy_target_group(proxied_group_id=ptg1['id'],
                                        expected_res_status=201)

    def test_proxy_group_no_update(self):
        ptg0 = self.create_policy_target_group()['policy_target_group']
        ptg1 = self.create_policy_target_group()['policy_target_group']
        ptg_proxy = self.create_policy_target_group(
            proxied_group_id=ptg0['id'])['policy_target_group']
        self.update_policy_target_group(
            ptg_proxy['id'], proxied_group_id=ptg1['id'],
            expected_res_status=400)

    def test_different_proxy_type(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        ptg_proxy = self.create_policy_target_group(
            proxied_group_id=ptg['id'], proxy_type='l2')['policy_target_group']
        self.assertEqual('l2', ptg_proxy['proxy_type'])

        ptg_proxy = self.show_policy_target_group(
            ptg_proxy['id'])['policy_target_group']
        self.assertEqual('l2', ptg_proxy['proxy_type'])

    def test_proxy_type_fails(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        res = self.create_policy_target_group(proxy_type='l2',
                                              expected_res_status=400)
        self.assertEqual('ProxyTypeSetWithoutProxiedPTG',
                         res['NeutronError']['type'])

        self.create_policy_target_group(proxied_group_id=ptg['id'],
                                        proxy_type='notvalid',
                                        expected_res_status=400)

    def test_proxy_gateway_no_proxy(self):
        ptg = self.create_policy_target_group()['policy_target_group']
        res = self.create_policy_target(
            policy_target_group_id=ptg['id'], proxy_gateway=True,
            expected_res_status=400)
        self.assertEqual('InvalidProxyGatewayGroup',
                         res['NeutronError']['type'])

    def test_proxy_pool_invalid_prefix_length(self):
        l3p = self.create_l3_policy(proxy_subnet_prefix_length=29)['l3_policy']
        res = self.update_l3_policy(l3p['id'], proxy_subnet_prefix_length=32,
                                    expected_res_status=400)
        self.assertEqual('InvalidDefaultSubnetPrefixLength',
                         res['NeutronError']['type'])

        # Verify change didn't persist
        l3p = self.show_l3_policy(l3p['id'])['l3_policy']
        self.assertEqual(29, l3p['proxy_subnet_prefix_length'])

        # Verify it fails in creation
        res = self.create_l3_policy(
            proxy_subnet_prefix_length=32, expected_res_status=400)
        self.assertEqual('InvalidDefaultSubnetPrefixLength',
                         res['NeutronError']['type'])

    def test_proxy_pool_legacy_invalid_version(self):
        cfg.CONF.set_override('use_subnetpools',
                              False,
                              group='resource_mapping')

        # without subnetpools, proxy_ip_pool of a different version
        # should fail
        res = self.create_l3_policy(ip_version=6, ip_pool='1::1/16',
                                    proxy_ip_pool='192.168.0.0/16',
                                    expected_res_status=400)
        self.assertEqual('InvalidIpPoolVersion', res['NeutronError']['type'])


class ExtensionDriverTestCase(test_ext_base.ExtensionDriverTestBase,
                              ExtensionDriverTestCaseMixin):
    _extension_drivers = ['proxy_group']
    _extension_path = None



# Copyright (c) 2019 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import defaultdict
from collections import namedtuple

import netaddr
from neutron.db.extra_dhcp_opt import models as dhcp_models
from neutron.db.models import allowed_address_pair as aap_models
from neutron.db.models import dns as dns_models
from neutron.db.models import l3 as l3_models
from neutron.db.models import securitygroup as sg_models
from neutron.db.models import segment as segment_models
from neutron.db import models_v2
from neutron.db.port_security import models as psec_models
from neutron.plugins.ml2 import models as ml2_models
from neutron.services.trunk import models as trunk_models
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_constants
from neutron_lib import context as n_context
from neutron_lib import rpc as n_rpc
from opflexagent import host_agent_rpc as oa_rpc
from opflexagent import rpc as o_rpc
from oslo_log import log
import oslo_messaging
from oslo_serialization import jsonutils
import sqlalchemy as sa
from sqlalchemy.ext import baked

from gbpservice.neutron.db import api as db_api
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import constants
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import db
from gbpservice.neutron.plugins.ml2plus.drivers.apic_aim import extension_db

LOG = log.getLogger(__name__)

BAKERY = baked.bakery(_size_alert=lambda c: LOG.warning(
    "sqlalchemy baked query cache size exceeded in %s", __name__))

EndpointPortInfo = namedtuple(
    'EndpointPortInfo',
    ['project_id',
     'port_id',
     'port_name',
     'network_id',
     'mac_address',
     'admin_state_up',
     'device_id',
     'device_owner',
     'host',
     'vif_type',
     'vif_details',
     'psec_enabled',
     'trunk_id',
     'subport_trunk_id',
     'net_mtu',
     'net_dns_domain',
     'nested_domain_name',
     'nested_domain_type',
     'nested_domain_infra_vlan',
     'nested_domain_service_vlan',
     'nested_domain_node_network_vlan',
     'svi',
     'epg_name',
     'epg_app_profile_name',
     'epg_tenant_name',
     'vrf_name',
     'vrf_tenant_name',
     'vm_name'])

EndpointFixedIpInfo = namedtuple(
    'EndpointFixedIpInfo',
    ['ip_address',
     'subnet_id',
     'ip_version',
     'cidr',
     'gateway_ip',
     'enable_dhcp',
     'dns_nameserver',
     'route_destination',
     'route_nexthop'])

EndpointBindingInfo = namedtuple(
    'EndpointBindingInfo',
    ['host',
     'level',
     'network_type',
     'physical_network',
     'segmentation_id'])

EndpointSecurityGroupInfo = namedtuple(
    'EndpointSecurityGroupInfo',
    ['sg_id',
     'project_id'])

EndpointDhcpIpInfo = namedtuple(
    'EndpointDhcpIpInfo',
    ['mac_address',
     'ip_address',
     'subnet_id'])

EndpointAapInfo = namedtuple(
    'EndpointAapInfo',
    ['mac_address',
     'ip_address'])

EndpointOwnedIpInfo = namedtuple(
    'EndpointOwnedIpInfo',
    ['ip_address',
     'actual_port_id'])

EndpointExternalNetworkInfo = namedtuple(
    'EndpointExternalNetworkInfo',
    ['network_id',
     'project_id',
     'epg_name',
     'epg_app_profile_name',
     'epg_tenant_name',
     'external_network_dn',
     'nat_type'])

EndpointFipInfo = namedtuple(
    'EndpointFipInfo',
    ['floating_ip_id',
     'floating_ip_address',
     'floating_network_id',
     'fixed_ip_address'])

EndpointSnatInfo = namedtuple(
    'EndpointSnatInfo',
    ['network_id',
     'ip_address',
     'cidr',
     'gateway_ip'])

EndpointTrunkInfo = namedtuple(
    'EndpointTrunkInfo',
    ['master_port_id',
     'subport_port_id',
     'segmentation_type',
     'segmentation_id'])


class TopologyRpcEndpoint(object):

    target = oslo_messaging.Target(version=oa_rpc.VERSION)

    def __init__(self, mechanism_driver):
        self.md = mechanism_driver

    @db_api.retry_if_session_inactive()
    def update_link(self, context, *args, **kwargs):
        context._session = db_api.get_writer_session()
        return self.md.update_link(context, *args, **kwargs)

    @db_api.retry_if_session_inactive()
    def delete_link(self, context, *args, **kwargs):
        # Don't take any action on link deletion in order to tolerate
        # situations like fabric upgrade or flapping links. Old links
        # are removed once a specific host is attached somewhere else.
        # To completely decommission the host, aimctl can be used to
        # cleanup the hostlink table.
        return


class ApicRpcHandlerMixin(object):

    def _start_rpc_listeners(self):
        conn = n_rpc.Connection()

        # Opflex RPC handler.
        self._opflex_endpoint = o_rpc.GBPServerRpcCallback(
            self, self.notifier)
        conn.create_consumer(
            o_rpc.TOPIC_OPFLEX,
            [self._opflex_endpoint],
            fanout=False)

        # Topology RPC hander.
        self._topology_endpoint = TopologyRpcEndpoint(self)
        conn.create_consumer(
            oa_rpc.TOPIC_APIC_SERVICE,
            [self._topology_endpoint],
            fanout=False)

        # Start listeners and return list of servers.
        return conn.consume_in_threads()

    # The following five methods handle RPCs from the Opflex agent.

    def get_gbp_details(self, context, **kwargs):
        LOG.debug("APIC AIM MD handling get_gbp_details for: %s", kwargs)

        # REVISIT: This RPC is no longer invoked by the Opflex agent,
        # and should be eliminated or should simply log an error, but
        # it is used extensively in unit tests.

        request = {'device': kwargs.get('device')}
        host = kwargs.get('host')
        response = self.request_endpoint_details(
            context, request=request, host=host)
        gbp_details = response.get('gbp_details')
        return gbp_details or response

    def get_vrf_details(self, context, **kwargs):
        LOG.debug("APIC AIM MD handling get_vrf_details for: %s", kwargs)

        vrf_id = kwargs.get('vrf_id')
        if not vrf_id:
            LOG.error("Missing vrf_id in get_vrf_details RPC: %s",
                      kwargs)
            return

        try:
            return self._get_vrf_details(context, vrf_id)
        except Exception as e:
            LOG.error("An exception occurred while processing "
                      "get_vrf_details RPC: %s", kwargs)
            LOG.exception(e)
            return {'l3_policy_id': vrf_id}

    def request_endpoint_details(self, context, **kwargs):
        LOG.debug("APIC AIM MD handling request_endpoint_details for: %s",
                  kwargs)

        request = kwargs.get('request')
        if not request:
            LOG.error("Missing request in request_endpoint_details RPC: %s",
                      kwargs)
            return

        device = request.get('device')
        if not device:
            LOG.error("Missing device in request_endpoint_details RPC: %s",
                      kwargs)
            return

        host = kwargs.get('host')
        if not host:
            LOG.error("Missing host in request_endpoint_details RPC: %s",
                      kwargs)
            return

        try:
            return self._request_endpoint_details(context, request, host)
        except Exception as e:
            LOG.error("An exception occurred while processing "
                      "request_endpoint_details RPC: %s", kwargs)
            LOG.exception(e)
            return {'device': device}

    def request_vrf_details(self, context, **kwargs):
        LOG.debug("APIC AIM MD handling request_vrf_details for: %s", kwargs)

        # REVISIT: This RPC is not currently invoked by the Opflex
        # agent, but that may be planned. Once it is, move the handler
        # implementation from get_vrf_details() to this method.
        return self.get_vrf_details(context, **kwargs)

    def ip_address_owner_update(self, context, **kwargs):
        LOG.debug("APIC AIM MD handling ip_address_owner_update for: %s",
                  kwargs)
        if not kwargs.get('ip_owner_info'):
            return
        ports_to_update = self.update_ip_owner(kwargs['ip_owner_info'])
        for p in ports_to_update:
            LOG.debug("APIC ownership update for port %s", p)
            self._notify_port_update(context, p)

    @db_api.retry_if_session_inactive()
    def _get_vrf_details(self, context, vrf_id):
        vrf_tenant_name, vrf_name = vrf_id.split(' ')
        with db_api.CONTEXT_READER.using(context) as session:
            vrf_subnets = self._query_vrf_subnets(
                session, vrf_tenant_name, vrf_name)
            return {
                'l3_policy_id': vrf_id,
                'vrf_tenant': vrf_tenant_name,
                'vrf_name': vrf_name,
                'vrf_subnets': vrf_subnets
            }

    @db_api.retry_if_session_inactive()
    def _request_endpoint_details(self, context, request, host):
        device = request['device']
        info = {'device': device}
        response = {
            'device': device,
            'request_id': request.get('request_id'),
            'timestamp': request.get('timestamp')
        }

        # Loop so we can bind the port, if necessary, outside the
        # transaction in which we query the endpoint's state, and then
        # retry.
        while True:
            # Start a read-only transaction. Separate read-write
            # transactions will be used if needed to bind the port or
            # assign SNAT IPs.
            with db_api.CONTEXT_READER.using(context) as session:
                # Extract possibly truncated port ID from device.
                #
                # REVISIT: If device identifies the port by its MAC
                # address instead of its UUID, _device_to_port_id()
                # will query for the entire port DB object. So
                # consider not calling _device_to_port_id() and
                # instead removing any device prefix here and
                # conditionally filtering in
                # _query_endpoint_port_info() below on either the
                # port's UUID or its mac_address.
                port_id = self.plugin._device_to_port_id(context, device)

                # Query for all the needed scalar (non-list) state
                # associated with the port.
                port_infos = self._query_endpoint_port_info(session, port_id)
                if not port_infos:
                    LOG.info("Nonexistent port %s in requent_endpoint_details "
                             "RPC from host %s", port_id, host)
                    return response
                if len(port_infos) > 1:
                    LOG.info("Multiple ports start with %s in "
                             "requent_endpoint_details RPC from host %s",
                             port_id, host)
                    return response
                port_info = port_infos[0]
                info['port_info'] = port_info

                # If port is bound, check host and do remaining
                # queries.
                if port_info.vif_type not in [
                        portbindings.VIF_TYPE_UNBOUND,
                        portbindings.VIF_TYPE_BINDING_FAILED]:

                    # Check that port is bound to host making the RPC
                    # request.
                    if port_info.host != host:
                        LOG.warning("Port %s bound to host %s, but "
                                    "request_endpoint_details RPC made from "
                                    "host %s",
                                    port_info.port_id, port_info.host, host)
                        return response

                    # Query for all needed state associated with each
                    # of the port's static IPs.
                    info['ip_info'] = self._query_endpoint_fixed_ip_info(
                        session, port_info.port_id)

                    # Query for list of state associated with each of
                    # the port's binding levels, sorted by level.
                    info['binding_info'] = self._query_endpoint_binding_info(
                        session, port_info.port_id)

                    # Query for list of the port's security groups.
                    info['sg_info'] = self._query_endpoint_sg_info(
                        session, port_info.port_id)

                    # Query for list of state associated with each
                    # DHCP IP on the port's network.
                    info['dhcp_ip_info'] = self._query_endpoint_dhcp_ip_info(
                        session, port_info.network_id)

                    # Query for the port's allowed address pairs.
                    info['aap_info'] = self._query_endpoint_aap_info(
                        session, port_info.port_id)

                    # Query for list of state associated with each of
                    # the port's HAIP owned IP addresses.
                    info['owned_ip_info'] = (
                        self._query_endpoint_haip_owned_ip_info(
                            session, port_info.port_id, port_info.network_id))

                    # Query for dict of state associated with the
                    # external networks to which the port's subnets
                    # are routed.
                    subnet_ids = set([ip.subnet_id for ip in info['ip_info']])
                    info['ext_net_info'] = self._query_endpoint_ext_net_info(
                        session, subnet_ids)

                    # Query for the port's active active AAP mode.
                    info['active_active_aap'] = self._query_active_active_aap(
                        session, subnet_ids)

                    # Query for list of floating IPs for both this
                    # port and all the other ports on which this
                    # port's HAIP owned addresses are actually
                    # defined.
                    fip_port_ids = (
                        [port_info.port_id] +
                        [x.actual_port_id for x in info['owned_ip_info']])
                    info['fip_info'] = self._query_endpoint_fip_info(
                        session, fip_port_ids)

                    # Query for dict of state associated with the SNAT
                    # ports on this host of the endpoint port's
                    # external networks.
                    info['snat_info'] = self._query_endpoint_snat_info(
                        session, host, list(info['ext_net_info'].keys()))

                    # Query for list of trunk subports for a trunk
                    # that the endpoint's port is associated with,
                    # either as the master port or as a subport.
                    trunk_id = port_info.trunk_id or port_info.subport_trunk_id
                    if trunk_id:
                        info['trunk_info'] = self._query_endpoint_trunk_info(
                            session, trunk_id)

                    # Query for the port's extra DHCP options.
                    info['extra_dhcp_opts'] = (
                        self._query_endpoint_extra_dhcp_opts(
                            session, port_info.port_id))

                    # Query for nested domain allowed VLANs for the
                    # port's network.
                    info['nested_domain_allowed_vlans'] = (
                        self._query_endpoint_nested_domain_allowed_vlans(
                            session, port_info.network_id))

                    # Query for VRF subnets.
                    info['vrf_subnets'] = self._query_vrf_subnets(
                        session, port_info.vrf_tenant_name, port_info.vrf_name)

                    # Let the GBP policy driver do its queries and add
                    # its info.
                    if self.gbp_driver:
                        self.gbp_driver.query_endpoint_rpc_info(session, info)

                    # Done with queries, so exit transaction and retry loop.
                    break

            # Attempt to bind port outside transaction.
            pc = self.plugin.get_bound_port_context(context, port_id, host)
            if (pc.vif_type == portbindings.VIF_TYPE_BINDING_FAILED or
                pc.vif_type == portbindings.VIF_TYPE_UNBOUND):
                LOG.warning("The request_endpoint_details RPC handler is "
                            "unable to bind port %s on host %s",
                            port_id, pc.host)
                return response

            # Successfully bound port, so loop to retry queries.

        # Completed queries, so build up the response.
        response['neutron_details'] = self._build_endpoint_neutron_details(
            info)
        response['gbp_details'] = self._build_endpoint_gbp_details(
            info, response['neutron_details'])
        response['trunk_details'] = self._build_endpoint_trunk_details(info)

        #  Let the GBP policy driver add/update its details in the response.
        if self.gbp_driver:
            self.gbp_driver.update_endpoint_rpc_details(info, response)

        # Return the response.
        return response

    def _query_endpoint_port_info(self, session, port_id):
        query = BAKERY(lambda s: s.query(
            models_v2.Port.project_id,
            models_v2.Port.id,
            models_v2.Port.name,
            models_v2.Port.network_id,
            models_v2.Port.mac_address,
            models_v2.Port.admin_state_up,
            models_v2.Port.device_id,
            models_v2.Port.device_owner,
            ml2_models.PortBinding.host,
            ml2_models.PortBinding.vif_type,
            ml2_models.PortBinding.vif_details,
            psec_models.PortSecurityBinding.port_security_enabled,
            trunk_models.Trunk.id,
            trunk_models.SubPort.trunk_id,
            models_v2.Network.mtu,
            dns_models.NetworkDNSDomain.dns_domain,
            extension_db.NetworkExtensionDb.nested_domain_name,
            extension_db.NetworkExtensionDb.nested_domain_type,
            extension_db.NetworkExtensionDb.nested_domain_infra_vlan,
            extension_db.NetworkExtensionDb.nested_domain_service_vlan,
            extension_db.NetworkExtensionDb.
            nested_domain_node_network_vlan,
            extension_db.NetworkExtensionDb.svi,
            db.NetworkMapping.epg_name,
            db.NetworkMapping.epg_app_profile_name,
            db.NetworkMapping.epg_tenant_name,
            db.NetworkMapping.vrf_name,
            db.NetworkMapping.vrf_tenant_name,
            db.VMName.vm_name,
        ))
        query += lambda q: q.outerjoin(
            ml2_models.PortBinding,
            ml2_models.PortBinding.port_id == models_v2.Port.id)
        query += lambda q: q.outerjoin(
            psec_models.PortSecurityBinding,
            psec_models.PortSecurityBinding.port_id == models_v2.Port.id)
        query += lambda q: q.outerjoin(
            trunk_models.Trunk,
            trunk_models.Trunk.port_id == models_v2.Port.id)
        query += lambda q: q.outerjoin(
            trunk_models.SubPort,
            trunk_models.SubPort.port_id == models_v2.Port.id)
        query += lambda q: q.outerjoin(
            models_v2.Network,
            models_v2.Network.id == models_v2.Port.network_id)
        query += lambda q: q.outerjoin(
            dns_models.NetworkDNSDomain,
            dns_models.NetworkDNSDomain.network_id ==
            models_v2.Port.network_id)
        query += lambda q: q.outerjoin(
            extension_db.NetworkExtensionDb,
            extension_db.NetworkExtensionDb.network_id ==
            models_v2.Port.network_id)
        query += lambda q: q.outerjoin(
            db.NetworkMapping,
            db.NetworkMapping.network_id == models_v2.Port.network_id)
        query += lambda q: q.outerjoin(
            db.VMName,
            db.VMName.device_id == models_v2.Port.device_id)
        query += lambda q: q.filter(
            models_v2.Port.id.startswith(sa.bindparam('port_id')))
        return [EndpointPortInfo._make(row) for row in
                query(session).params(
                    port_id=port_id)]

    def _query_endpoint_fixed_ip_info(self, session, port_id):
        # In this query, IPAllocations are outerjoined with
        # DNSNameServers and SubnetRoutes. This avoids needing to make
        # separate queries for DNSNameServers and for SubnetRoutes,
        # but results in rows being returned for the cross product of
        # the DNSNameServer rows and SubnetRoute rows associated with
        # each fixed IP. Unless there are use cases where large
        # numbers of rows in both these tables exist for the same
        # fixed IP, this approach is expected to provide better
        # latency and scalability than using separate
        # queries. Redundant information must be ignored when
        # processing the rows returned from this query.
        query = BAKERY(lambda s: s.query(
            models_v2.IPAllocation.ip_address,
            models_v2.IPAllocation.subnet_id,
            models_v2.Subnet.ip_version,
            models_v2.Subnet.cidr,
            models_v2.Subnet.gateway_ip,
            models_v2.Subnet.enable_dhcp,
            models_v2.DNSNameServer.address,
            models_v2.SubnetRoute.destination,
            models_v2.SubnetRoute.nexthop,
        ))
        query += lambda q: q.join(
            models_v2.Subnet,
            models_v2.Subnet.id == models_v2.IPAllocation.subnet_id)
        query += lambda q: q.outerjoin(
            models_v2.DNSNameServer,
            models_v2.DNSNameServer.subnet_id ==
            models_v2.IPAllocation.subnet_id)
        query += lambda q: q.outerjoin(
            models_v2.SubnetRoute,
            models_v2.SubnetRoute.subnet_id ==
            models_v2.IPAllocation.subnet_id)
        query += lambda q: q.filter(
            models_v2.IPAllocation.port_id == sa.bindparam('port_id'))
        query += lambda q: q.order_by(
            models_v2.DNSNameServer.order)
        return [EndpointFixedIpInfo._make(row) for row in
                query(session).params(
                    port_id=port_id)]

    def _query_endpoint_binding_info(self, session, port_id):
        query = BAKERY(lambda s: s.query(
            ml2_models.PortBindingLevel.host,
            ml2_models.PortBindingLevel.level,
            segment_models.NetworkSegment.network_type,
            segment_models.NetworkSegment.physical_network,
            segment_models.NetworkSegment.segmentation_id,
        ))
        query += lambda q: q.join(
            segment_models.NetworkSegment,
            segment_models.NetworkSegment.id ==
            ml2_models.PortBindingLevel.segment_id)
        query += lambda q: q.filter(
            ml2_models.PortBindingLevel.port_id == sa.bindparam('port_id'))
        query += lambda q: q.order_by(
            ml2_models.PortBindingLevel.level)
        return [EndpointBindingInfo._make(row) for row in
                query(session).params(
                    port_id=port_id)]

    def _query_endpoint_sg_info(self, session, port_id):
        query = BAKERY(lambda s: s.query(
            sg_models.SecurityGroup.id,
            sg_models.SecurityGroup.project_id,
        ))
        query += lambda q: q.join(
            sg_models.SecurityGroupPortBinding,
            sg_models.SecurityGroupPortBinding.security_group_id ==
            sg_models.SecurityGroup.id)
        query += lambda q: q.filter(
            sg_models.SecurityGroupPortBinding.port_id ==
            sa.bindparam('port_id'))
        return [EndpointSecurityGroupInfo._make(row) for row in
                query(session).params(
                    port_id=port_id)]

    def _query_endpoint_dhcp_ip_info(self, session, network_id):
        query = BAKERY(lambda s: s.query(
            models_v2.Port.mac_address,
            models_v2.IPAllocation.ip_address,
            models_v2.IPAllocation.subnet_id,
        ))
        query += lambda q: q.join(
            models_v2.IPAllocation,
            models_v2.IPAllocation.port_id == models_v2.Port.id)
        query += lambda q: q.filter(
            models_v2.Port.network_id == sa.bindparam('network_id'),
            models_v2.Port.device_owner == n_constants.DEVICE_OWNER_DHCP)
        return [EndpointDhcpIpInfo._make(row) for row in
                query(session).params(
                    network_id=network_id)]

    def _query_endpoint_aap_info(self, session, port_id):
        query = BAKERY(lambda s: s.query(
            aap_models.AllowedAddressPair.mac_address,
            aap_models.AllowedAddressPair.ip_address,
        ))
        query += lambda q: q.filter(
            aap_models.AllowedAddressPair.port_id ==
            sa.bindparam('port_id'))
        return [EndpointAapInfo._make(row) for row in
                query(session).params(
                    port_id=port_id)]

    def _query_endpoint_haip_owned_ip_info(self, session, port_id, network_id):
        query = BAKERY(lambda s: s.query(
            db.HAIPAddressToPortAssociation.ha_ip_address,
            models_v2.IPAllocation.port_id,
        ))
        query += lambda q: q.outerjoin(
            models_v2.IPAllocation,
            models_v2.IPAllocation.ip_address ==
            db.HAIPAddressToPortAssociation.ha_ip_address)
        query += lambda q: q.filter(
            db.HAIPAddressToPortAssociation.port_id ==
            sa.bindparam('port_id'),
            sa.or_(models_v2.IPAllocation.network_id.is_(None),
                   models_v2.IPAllocation.network_id ==
                   sa.bindparam('network_id')))
        return [EndpointOwnedIpInfo._make(row) for row in
                query(session).params(
                    port_id=port_id,
                    network_id=network_id)]

    def _query_active_active_aap(self, session, subnet_ids):
        # Default is False.
        if not subnet_ids:
            return False
        query = BAKERY(lambda s: s.query(
            extension_db.SubnetExtensionDb.active_active_aap,
        ))
        query += lambda q: q.filter(
            extension_db.SubnetExtensionDb.subnet_id.in_(
                sa.bindparam('subnet_ids', expanding=True)))
        query += lambda q: q.distinct()
        for active_active_aap, in query(session).params(
                subnet_ids=list(subnet_ids)):
            if active_active_aap is False:
                return False
        # Return True only if all the subnets have this flag enabled.
        return True

    def _query_endpoint_ext_net_info(self, session, subnet_ids):
        # REVISIT: Consider replacing this query with additional joins
        # in _query_endpoint_fixed_ip_info to eliminate a round-trip
        # to the DB server. This would require using aliases to
        # disambiguate between the endpoint's port's IPAllocation and
        # the router port's IPAllocation, and its not obvious if
        # aliases can be used with baked queries.
        if not subnet_ids:
            return {}
        query = BAKERY(lambda s: s.query(
            models_v2.Network.id,
            models_v2.Network.project_id,
            db.NetworkMapping.epg_name,
            db.NetworkMapping.epg_app_profile_name,
            db.NetworkMapping.epg_tenant_name,
            extension_db.NetworkExtensionDb.external_network_dn,
            extension_db.NetworkExtensionDb.nat_type,
        ))
        query += lambda q: q.join(
            models_v2.Port,  # router's gw_port
            models_v2.Port.network_id == models_v2.Network.id)
        query += lambda q: q.join(
            l3_models.Router,
            l3_models.Router.gw_port_id == models_v2.Port.id)
        query += lambda q: q.join(
            l3_models.RouterPort,
            l3_models.RouterPort.router_id == l3_models.Router.id and
            l3_models.RouterPort.port_type ==
            n_constants.DEVICE_OWNER_ROUTER_INTF)
        query += lambda q: q.join(
            models_v2.IPAllocation,  # router interface IP
            models_v2.IPAllocation.port_id == l3_models.RouterPort.port_id)
        query += lambda q: q.join(
            db.NetworkMapping,  # mapping of gw_port's network
            db.NetworkMapping.network_id == models_v2.Port.network_id)
        query += lambda q: q.outerjoin(
            extension_db.NetworkExtensionDb,
            extension_db.NetworkExtensionDb.network_id ==
            models_v2.Port.network_id)
        query += lambda q: q.filter(
            models_v2.IPAllocation.subnet_id.in_(
                sa.bindparam('subnet_ids', expanding=True)))
        query += lambda q: q.distinct()
        return {row[0]: EndpointExternalNetworkInfo._make(row) for row in
                query(session).params(
                    subnet_ids=list(subnet_ids))}

    def _query_endpoint_fip_info(self, session, port_ids):
        if not port_ids:
            return []
        query = BAKERY(lambda s: s.query(
            l3_models.FloatingIP.id,
            l3_models.FloatingIP.floating_ip_address,
            l3_models.FloatingIP.floating_network_id,
            l3_models.FloatingIP.fixed_ip_address,
        ))
        query += lambda q: q.filter(
            l3_models.FloatingIP.fixed_port_id.in_(sa.bindparam(
                'port_ids', expanding=True)))
        return [EndpointFipInfo._make(row) for row in
                query(session).params(
                    port_ids=port_ids)]

    def _query_endpoint_snat_info(self, session, host, ext_net_ids):
        # REVISIT: Consider replacing this query with additional joins
        # in _query_endpoint_ext_net_info to eliminate a round-trip to
        # the DB server. This would require using aliases to
        # disambiguate tables appearing multiple times in the query,
        # and its not obvious if aliases can be used with baked
        # queries.
        if not ext_net_ids:
            return {}
        query = BAKERY(lambda s: s.query(
            models_v2.Port.network_id,
            models_v2.IPAllocation.ip_address,
            models_v2.Subnet.cidr,
            models_v2.Subnet.gateway_ip,
        ))
        query += lambda q: q.join(
            models_v2.IPAllocation,
            models_v2.IPAllocation.port_id == models_v2.Port.id)
        query += lambda q: q.join(
            models_v2.Subnet,
            models_v2.Subnet.id == models_v2.IPAllocation.subnet_id)
        query += lambda q: q.filter(
            models_v2.Port.network_id.in_(sa.bindparam(
                'ext_net_ids', expanding=True)),
            models_v2.Port.device_id == sa.bindparam('host'),
            models_v2.Port.device_owner == constants.DEVICE_OWNER_SNAT_PORT)
        return {row[0]: EndpointSnatInfo._make(row) for row in
                query(session).params(
                    host=host,
                    ext_net_ids=ext_net_ids)}

    def _query_endpoint_trunk_info(self, session, trunk_id):
        query = BAKERY(lambda s: s.query(
            trunk_models.Trunk.port_id,
            trunk_models.SubPort.port_id,
            trunk_models.SubPort.segmentation_type,
            trunk_models.SubPort.segmentation_id,
        ))
        query += lambda q: q.outerjoin(
            trunk_models.SubPort,
            trunk_models.SubPort.trunk_id == trunk_models.Trunk.id)
        query += lambda q: q.filter(
            trunk_models.Trunk.id == sa.bindparam('trunk_id'))
        return [EndpointTrunkInfo._make(row) for row in
                query(session).params(
                    trunk_id=trunk_id)]

    def _query_endpoint_extra_dhcp_opts(self, session, port_id):
        query = BAKERY(lambda s: s.query(
            dhcp_models.ExtraDhcpOpt.opt_name,
            dhcp_models.ExtraDhcpOpt.opt_value,
        ))
        query += lambda q: q.filter(
            dhcp_models.ExtraDhcpOpt.port_id == sa.bindparam('port_id'))
        return {k: v for k, v in query(session).params(
            port_id=port_id)}

    def _query_endpoint_nested_domain_allowed_vlans(self, session, network_id):
        query = BAKERY(lambda s: s.query(
            extension_db.NetworkExtNestedDomainAllowedVlansDb.vlan,
        ))
        query += lambda q: q.filter(
            extension_db.NetworkExtNestedDomainAllowedVlansDb.network_id ==
            sa.bindparam('network_id'))
        return [x for x, in query(session).params(
            network_id=network_id)]

    def _query_vrf_subnets(self, session, vrf_tenant_name, vrf_name):
        # A VRF mapped from one or two (IPv4 and/or IPv6)
        # address_scopes cannot be associated with unscoped
        # subnets. So first see if the VRF is mapped from
        # address_scopes, and if so, return the subnetpool CIDRs
        # associated with those address_scopes.
        query = BAKERY(lambda s: s.query(
            models_v2.SubnetPoolPrefix.cidr))
        query += lambda q: q.join(
            models_v2.SubnetPool,
            models_v2.SubnetPool.id ==
            models_v2.SubnetPoolPrefix.subnetpool_id)
        query += lambda q: q.join(
            db.AddressScopeMapping,
            db.AddressScopeMapping.scope_id ==
            models_v2.SubnetPool.address_scope_id)
        query += lambda q: q.filter(
            db.AddressScopeMapping.vrf_name ==
            sa.bindparam('vrf_name'),
            db.AddressScopeMapping.vrf_tenant_name ==
            sa.bindparam('vrf_tenant_name'))
        result = [x for x, in query(session).params(
            vrf_name=vrf_name,
            vrf_tenant_name=vrf_tenant_name)]
        if result:
            return result

        # If the VRF is not mapped from address_scopes, return the
        # CIDRs of all the subnets on all the networks associated with
        # the VRF.
        #
        # REVISIT: Consider combining these two queries into a single
        # query, using outerjoins to SubnetPool and
        # AddressScopeMapping. But that would result in all the
        # subnets' CIDRs being returned, even for the scoped case
        # where they are not needed, so it may not be a win.
        query = BAKERY(lambda s: s.query(
            models_v2.Subnet.cidr))
        query += lambda q: q.join(
            db.NetworkMapping,
            db.NetworkMapping.network_id ==
            models_v2.Subnet.network_id)
        query += lambda q: q.filter(
            db.NetworkMapping.vrf_name ==
            sa.bindparam('vrf_name'),
            db.NetworkMapping.vrf_tenant_name ==
            sa.bindparam('vrf_tenant_name'))
        return [x for x, in query(session).params(
            vrf_name=vrf_name,
            vrf_tenant_name=vrf_tenant_name)]

    def _build_endpoint_neutron_details(self, info):
        port_info = info['port_info']
        binding_info = info['binding_info']

        details = {}
        details['admin_state_up'] = port_info.admin_state_up
        details['device_owner'] = port_info.device_owner
        details['fixed_ips'] = self._build_fixed_ips(info)
        details['network_id'] = port_info.network_id
        details['network_type'] = binding_info[-1].network_type
        details['physical_network'] = binding_info[-1].physical_network
        details['segmentation_id'] = binding_info[-1].segmentation_id
        details['port_id'] = port_info.port_id

        return details

    def _build_fixed_ips(self, info):
        ip_info = info['ip_info']

        # Build dict of unique fixed IPs, ignoring duplicates due to
        # joins between Port and DNSNameServers and Routes.
        fixed_ips = {}
        for ip in ip_info:
            if ip.ip_address not in fixed_ips:
                fixed_ips[ip.ip_address] = {'subnet_id': ip.subnet_id,
                                            'ip_address': ip.ip_address}

        return list(fixed_ips.values())

    def _build_endpoint_gbp_details(self, info, neutron_details):
        port_info = info['port_info']

        # Note that the GBP policy driver will replace these
        # app_profile_name, endpoint_group_name, ptg_tenant,
        # ... values if the port belongs to a GBP PolicyTarget.

        details = {}
        details['allowed_address_pairs'] = self._build_aaps(info)
        details['active_active_aap'] = info['active_active_aap']
        details['app_profile_name'] = port_info.epg_app_profile_name
        details['device'] = info['device']  # Redundant.
        if self.apic_optimized_dhcp_lease_time > 0:
            details['dhcp_lease_time'] = self.apic_optimized_dhcp_lease_time
        details['dns_domain'] = port_info.net_dns_domain or ''
        details['dns_domain'] = details['dns_domain'].rstrip('.')
        details['enable_dhcp_optimization'] = self.enable_dhcp_opt
        details['enable_metadata_optimization'] = self.enable_metadata_opt
        if port_info.svi:
            details['endpoint_group_name'] = port_info.network_id
            details['svi'] = True
        else:
            details['endpoint_group_name'] = port_info.epg_name
        details['floating_ip'] = self._build_fips(info)
        details['host'] = port_info.host
        details['host_snat_ips'] = self._build_host_snat_ips(
            info, details, neutron_details)
        mtu = self._get_interface_mtu(info)
        if mtu:
            details['interface_mtu'] = mtu
        details['ip_mapping'] = self._build_ipms(info, details)
        details['l3_policy_id'] = ("%s %s" %
                                   (port_info.vrf_tenant_name,
                                    port_info.vrf_name))
        details['mac_address'] = port_info.mac_address
        details['nested_domain_allowed_vlans'] = (
            info['nested_domain_allowed_vlans'])
        details['nested_domain_infra_vlan'] = (
            port_info.nested_domain_infra_vlan)
        details['nested_domain_name'] = port_info.nested_domain_name
        details['nested_domain_node_network_vlan'] = (
            port_info.nested_domain_node_network_vlan)
        details['nested_domain_service_vlan'] = (
            port_info.nested_domain_service_vlan)
        details['nested_domain_type'] = port_info.nested_domain_type
        details['nested_host_vlan'] = (
            self.nested_host_vlan if port_info.nested_domain_infra_vlan
            else None)
        details['port_id'] = port_info.port_id  # Redundant.
        details['promiscuous_mode'] = self._get_promiscuous_mode(info)
        details['ptg_tenant'] = port_info.epg_tenant_name
        if info['sg_info']:
            # Only add security group details if the port has SGs and
            # it doesn't belong to a legacy VM using iptables.
            vif_details = (port_info.vif_details and
                           jsonutils.loads(port_info.vif_details))
            if not (vif_details and vif_details.get('port_filter') and
                    vif_details.get('ovs_hybrid_plug')):
                details['security_group'] = self._build_sg_details(info)
        details['subnets'] = self._build_subnet_details(info)
        details['vm-name'] = (port_info.vm_name if
                              port_info.device_owner.startswith('compute:') and
                              port_info.vm_name else port_info.device_id)
        details['vrf_name'] = port_info.vrf_name
        details['vrf_subnets'] = info['vrf_subnets']
        details['vrf_tenant'] = port_info.vrf_tenant_name

        return details

    def _build_aaps(self, info):
        # If active_active_app is enabled, then the concept of
        # owned_ip doesn't apply.
        if info['active_active_aap']:
            owned_ips = set()
        else:
            owned_ips = set(ip.ip_address for ip in info['owned_ip_info'])
        aaps = {}
        for allowed in info['aap_info']:
            aaps[allowed.ip_address] = {'ip_address': allowed.ip_address,
                                        'mac_address': allowed.mac_address}
            cidr = netaddr.IPNetwork(allowed.ip_address)
            if ((cidr.version == 4 and cidr.prefixlen != 32) or
                (cidr.version == 6 and cidr.prefixlen != 128)):
                # Never mark CIDRs as "active", but
                # look for owned addresses in this CIDR, and
                # if present, add them to the allowed-address-pairs
                # list, and mark those as "active".
                for ip in owned_ips:
                    if ip in cidr and ip not in aaps:
                        aaps[ip] = {'ip_address': ip,
                                    'mac_address': allowed.mac_address,
                                    'active': True}
            elif allowed.ip_address in owned_ips:
                aaps[allowed.ip_address]['active'] = True
        return list(aaps.values())

    def _build_fips(self, info):
        ext_net_info = info['ext_net_info']
        fips = []
        for fip in info['fip_info']:
            details = {'id': fip.floating_ip_id,
                       'fixed_ip_address': fip.fixed_ip_address,
                       'floating_ip_address': fip.floating_ip_address}
            ext_net = ext_net_info.get(fip.floating_network_id)
            if (ext_net and ext_net.external_network_dn and
                ext_net.nat_type == 'distributed'):
                details['nat_epg_app_profile'] = ext_net.epg_app_profile_name
                details['nat_epg_name'] = ext_net.epg_name
                details['nat_epg_tenant'] = ext_net.epg_tenant_name
                details['ext_net_id'] = fip.floating_network_id
            fips.append(details)
        return fips

    def _build_host_snat_ips(self, info, details, neutron_details):
        snat_info = info['snat_info']
        host = info['port_info'].host

        fip_fixed_ips = {}
        for fip in details['floating_ip']:
            if 'ext_net_id' in fip:
                fip_fixed_ips.setdefault(fip['ext_net_id'], set()).add(
                    fip['fixed_ip_address'])

        ips = [x['ip_address'] for x in neutron_details['fixed_ips']]
        ips_aap = [aap['ip_address'] for aap in
                   details['allowed_address_pairs'] if aap.get('active',
                                                               False)]
        total_ips = sorted(ips + ips_aap)

        host_snat_ips = []
        for ext_net in info['ext_net_info'].values():
            need_snat = False
            for ip in total_ips:
                if ip not in fip_fixed_ips.get(ext_net.network_id, []):
                    need_snat = True
                    break
            if need_snat is False:
                continue

            snat = snat_info.get(ext_net.network_id)
            if snat:
                snat_ip = {'host_snat_ip': snat.ip_address,
                           'gateway_ip': snat.gateway_ip,
                           'prefixlen': int(snat.cidr.split('/')[1])}
            else:
                # No existing SNAT IP for this external network on
                # this host, so allocate one.
                ctx = n_context.get_admin_context()
                snat_ip = self.get_or_allocate_snat_ip(
                    ctx, host, {'id': ext_net.network_id,
                                'tenant_id': ext_net.project_id})
            if snat_ip:
                snat_ip['ext_net_id'] = ext_net.network_id
                snat_ip['external_segment_name'] = (
                    ext_net.external_network_dn.replace('/', ':'))
                host_snat_ips.append(snat_ip)
        return host_snat_ips

    def _get_interface_mtu(self, info):
        if self.advertise_mtu:
            opts = info['extra_dhcp_opts']
            opt_value = opts.get('interface-mtu') or opts.get('26')
            if opt_value:
                try:
                    return int(opt_value)
                except ValueError:
                    pass
            return info['port_info'].net_mtu

    def _build_ipms(self, info, details):
        host_snat_ips = details['host_snat_ips']
        host_snat_ext_net_ids = [host_snat_ip['ext_net_id'] for
                                 host_snat_ip in host_snat_ips]
        return [{'external_segment_name':
                 ext_net.external_network_dn.replace('/', ':'),
                 'nat_epg_app_profile': ext_net.epg_app_profile_name,
                 'nat_epg_name': ext_net.epg_name,
                 'nat_epg_tenant': ext_net.epg_tenant_name}
                for ext_net in info['ext_net_info'].values()
                if ext_net.external_network_dn and
                ext_net.nat_type == 'distributed' and
                ext_net.network_id in host_snat_ext_net_ids]

    def _get_promiscuous_mode(self, info):
        port_info = info['port_info']
        # REVISIT: Replace PROMISCUOUS_SUFFIX with a proper API
        # attribute if really needed, but why not just have
        # applications use port_security_enabled=False?
        return (port_info.device_owner in constants.PROMISCUOUS_TYPES or
                port_info.port_name.endswith(constants.PROMISCUOUS_SUFFIX) or
                port_info.psec_enabled is False)

    def _build_sg_details(self, info):
        return (
            [{'policy-space': self.name_mapper.project(None, sg.project_id),
              'name': sg.sg_id} for sg in info['sg_info']] +
            [{'policy-space': 'common', 'name': self._default_sg_name}])

    def _build_subnet_details(self, info):
        ip_info = info['ip_info']
        dhcp_ip_info = info['dhcp_ip_info']

        # Build dict of subnets with basic subnet details, and collect
        # joined DNSNameServer and Route info. Order must be preserved
        # among DNSNameServer entries for a subnet.
        subnets = {}
        subnet_dns_nameservers = defaultdict(list)
        subnet_routes = defaultdict(set)
        for ip in ip_info:
            if ip.subnet_id not in subnets:
                subnet = {}
                subnet['cidr'] = ip.cidr
                subnet['enable_dhcp'] = ip.enable_dhcp
                subnet['gateway_ip'] = ip.gateway_ip
                subnet['id'] = ip.subnet_id
                subnet['ip_version'] = ip.ip_version
                subnets[ip.subnet_id] = subnet
            if ip.dns_nameserver:
                dns_nameservers = subnet_dns_nameservers[ip.subnet_id]
                if ip.dns_nameserver not in dns_nameservers:
                    dns_nameservers.append(ip.dns_nameserver)
            if ip.route_destination:
                subnet_routes[ip.subnet_id].add(
                    (ip.route_destination, ip.route_nexthop))

        # Add remaining details to each subnet.
        for subnet_id, subnet in subnets.items():
            dhcp_ips = set()
            dhcp_ports = defaultdict(list)
            for ip in dhcp_ip_info:
                if ip.subnet_id == subnet_id:
                    dhcp_ips.add(ip.ip_address)
                    dhcp_ports[ip.mac_address].append(ip.ip_address)
            dhcp_ips = list(dhcp_ips)

            routes = subnet_routes[subnet_id]
            if subnet['ip_version'] == 4:
                # Find default and metadata routes.
                default_routes = set()
                metadata_routes = set()
                for route in routes:
                    destination = route[0]
                    if destination == constants.IPV4_ANY_CIDR:
                        default_routes.add(route)
                    elif destination == constants.IPV4_METADATA_CIDR:
                        metadata_routes.add(route)
                # Add gateway_ip and missing routes. Note that these
                # might get removed by the GBP PD if the L2P's
                # inject_default_route attribute is False.
                gateway_ip = subnet['gateway_ip']
                if not default_routes and gateway_ip:
                    routes.add((constants.IPV4_ANY_CIDR, gateway_ip))
                # REVISIT: We need to decide if we should provide
                # host-routes for all of the DHCP agents. For now
                # use the first DHCP agent in our list for the
                # metadata host-route next-hop IPs.
                if (not metadata_routes and dhcp_ports and
                    (not self.enable_metadata_opt or
                     (self.enable_metadata_opt and not default_routes))):
                    for ip in dhcp_ports[list(dhcp_ports.keys())[0]]:
                        routes.add((constants.IPV4_METADATA_CIDR, ip))

            subnet['dhcp_server_ips'] = dhcp_ips
            subnet['dhcp_server_ports'] = dhcp_ports
            subnet['dns_nameservers'] = (subnet_dns_nameservers[subnet_id] or
                                         dhcp_ips)
            subnet['host_routes'] = [
                {'destination': destination, 'nexthop': nexthop}
                for destination, nexthop in routes]

        return list(subnets.values())

    def _build_endpoint_trunk_details(self, info):
        trunk_info = info.get('trunk_info')
        if not trunk_info:
            return
        port_info = info.get('port_info')
        return {'trunk_id': port_info.trunk_id or port_info.subport_trunk_id,
                'master_port_id': trunk_info[0].master_port_id,
                'subports': [{'port_id': sp.subport_port_id,
                              'segmentation_type': sp.segmentation_type,
                              'segmentation_id': sp.segmentation_id}
                             for sp in trunk_info if (
                             sp.subport_port_id is not None)]}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory
from oslo_log import helpers as log

from gbpservice.neutron.services.grouppolicy.common import exceptions as exc
from gbpservice.neutron.services.grouppolicy.drivers import (
    implicit_policy as ipd)
from gbpservice.neutron.services.grouppolicy.drivers import (
    resource_mapping as rmd)


class CommonNeutronBase(ipd.ImplicitPolicyBase, rmd.OwnedResourcesOperations,
                        rmd.ImplicitResourceOperations):
    """Neutron Resources' Orchestration driver.

    This driver realizes GBP's network semantics by orchestrating
    the necessary Neutron resources.
    """

    @log.log_method_call
    def initialize(self):
        # REVISIT: Check if this is still required
        self._cached_agent_notifier = None
        self._gbp_plugin = None
        super(CommonNeutronBase, self).initialize()

    @property
    def gbp_plugin(self):
        if not self._gbp_plugin:
            self._gbp_plugin = directory.get_plugin("GROUP_POLICY")
        return self._gbp_plugin

    @log.log_method_call
    def create_l2_policy_postcommit(self, context):
        if not context.current['l3_policy_id']:
            self._use_implicit_l3_policy(context)
        l3p_db = context._plugin._get_l3_policy(
            context._plugin_context, context.current['l3_policy_id'])
        if not context.current['network_id']:
            self._use_implicit_network(
                context, address_scope_v4=l3p_db['address_scope_v4_id'],
                address_scope_v6=l3p_db['address_scope_v6_id'])

    @log.log_method_call
    def update_l2_policy_precommit(self, context):
        if (context.current['inject_default_route'] !=
            context.original['inject_default_route']):
            raise exc.UnsettingInjectDefaultRouteOfL2PolicyNotSupported()
        if (context.current['l3_policy_id'] !=
            context.original['l3_policy_id']):
            raise exc.L3PolicyUpdateOfL2PolicyNotSupported()

    @log.log_method_call
    def delete_l2_policy_postcommit(self, context):
        network_id = context.current['network_id']
        if network_id:
            self._cleanup_network(context._plugin_context, network_id)
        l3p_id = context.current['l3_policy_id']
        if l3p_id:
            self._cleanup_l3_policy(context, l3p_id)

    def _port_id_to_pt(self, plugin_context, port_id):
        pts = self.gbp_plugin.get_policy_targets(
            plugin_context, {'port_id': [port_id]})
        if pts:
            return pts[0]

    def _pt_to_ptg(self, plugin_context, pt):
        if pt:
            return self.gbp_plugin.get_policy_target_group(
                plugin_context, pt['policy_target_group_id'])
        return None

    def _port_id_to_ptg(self, plugin_context, port_id):
        pt = self._port_id_to_pt(plugin_context, port_id)
        return self._pt_to_ptg(plugin_context, pt), pt

    def _network_id_to_l2p(self, context, network_id):
        l2ps = self.gbp_plugin.get_l2_policies(
            context, filters={'network_id': [network_id]})
        for l2p in l2ps:
            if l2p['network_id'] == network_id:
                return l2p



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from gbpservice.neutron.db.grouppolicy.extensions import (
    apic_reuse_bd_db as db)
from gbpservice.neutron.db.grouppolicy import group_policy_db as gp_db
from gbpservice.neutron.extensions import apic_reuse_bd as ext
from gbpservice.neutron.extensions import group_policy as gpolicy
from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)

LOG = logging.getLogger(__name__)


class ApicReuseBdExtensionDriver(api.ExtensionDriver,
                                 db.ApicReuseBdDBMixin):
    _supported_extension_alias = ext.CISCO_APIC_GBP_REUSE_BD_EXT
    _extension_dict = ext.EXTENDED_ATTRIBUTES_2_0

    def __init__(self):
        LOG.info("ApicReuseBdExtensionDriver __init__")

    def initialize(self):
        pass

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_create_l2_policy(self, session, data, result):
        l2p = data['l2_policy']
        if l2p.get('reuse_bd'):
            target_l2p = (session.query(gp_db.L2Policy)
                          .filter_by(id=l2p['reuse_bd']).first())
            if not target_l2p:
                raise gpolicy.L2PolicyNotFound(l2_policy_id=l2p['reuse_bd'])
            self.add_reuse_bd_l2policy(session, result['id'], l2p['reuse_bd'])
            result['reuse_bd'] = l2p['reuse_bd']

    def extend_l2_policy_dict(self, session, result):
        row = self.get_reuse_bd_l2policy(session, l2_policy_id=result['id'])
        if row:
            result['reuse_bd'] = row.target_l2_policy_id



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import re
import subprocess
import sys

import paramiko
import prettytable
import six


def sshconnect(hostname, user, passwd):
    sshclient = paramiko.SSHClient()
    sshclient.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        sshclient.connect(hostname, username=user, password=passwd)
    except Exception:
        # raise ErrorConnectingToServer(
        #    "Error connecting to server %s: %s" %
        #    (hostname, e))
        # sshclient = None
        raise
    return sshclient


def report_table(suite_name):
    ps = subprocess.Popen(['grep',
                           '-r',
                           'TESTCASE',
                           '/tmp/%s.log' % (suite_name)],
                          stdout=subprocess.PIPE)
    output = ps.communicate()[0]
    output = output.splitlines()
    line = 0
    tc_dict = {}
    while line < len(output):
        find1 = re.search('\\b(TESTCASE_GBP_.*)\\b: (.*)', output[line], re.I)
        if find1 is not None:
            line += 1
            if line <= len(output) - 1:
                find2 = re.search(
                    '\\b%s\\b: (.*)' %
                    (find1.group(1)), output[line], re.I)
                if find2 is not None:
                    tc_dict[find1.group(1)] = find2.group(1), find1.group(2)
        line += 1
    table = prettytable.PrettyTable(["TESTCASE_ID", "RESULTS",
                                     "TESTCASE_HEADER"])
    table.padding_width = 1
    for key, val in six.iteritems(tc_dict):
        table.add_row(["%s" % (key), "%s" % (val[0]), "%s" % (val[1])])
    return table


def report_results(suite_name, txt_file):
    orig_stdout = sys.stdout
    f = open('%s' % (txt_file), 'a')
    sys.stdout = f
    report = report_table(suite_name)
    print(report)
    sys.stdout = orig_stdout
    f.close()


def gen_tc_header():
    comb_list = [
        ['same_leaf', 'two_leafs'], [
            'same_host', 'two_hosts'], [
                'same_ptg', 'two_ptgs'], [
                    'same_L3_subnet', 'two_L3_subnets'], [
                        'same_L2_bd', 'two_L2_bds']]

    out_hdr_list = list(itertools.product(*comb_list))
    out_headers = []
    for hdr in out_hdr_list:
        header = 'test_' + '_'.join(str(i) for i in hdr)
        out_headers.append(header)

    proto = list(itertools.combinations(
        ['icmp', 'tcp', 'udp', 'dhcp', 'arp'], 2))
    proto_hdrs = []
    for hdr in proto:
        proto_header = '_'.join(str(i) for i in hdr)
        proto_hdrs.append(proto_header)
    in_hdrs = list(itertools.product(out_headers, proto_hdrs))
    final_headers = []
    for hdr in in_hdrs:
        tc_header = '_'.join(str(i) for i in hdr)
        final_headers.append(tc_header)
    table = prettytable.PrettyTable(["TESTCASE_ID", "STATUS",
                                     "TESTCASE_HEADER"])
    table.padding_width = 1
    for i in range(len(final_headers)):
        table.add_row(["TESTCASE_DP_%s" %
                       (i + 1), "TBA", "%s" %
                       (final_headers[i])])
    print(table)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import commands
import logging
import os
import sys

from libs import config_libs
from libs import utils_libs
from libs import verify_libs


def main():

    # Run the Testcases:
    test = test_gbp_pr_neg()
    if test.test_gbp_pr_neg_1() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PR_NEG_1')
    if test.test_gbp_pr_neg_2() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PR_NEG_2')
    if test.test_gbp_pr_neg_3() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PR_NEG_3')
    if test.test_gbp_pr_neg_4() == 0:
        test.cleanup(tc_name='TESTCASE_GBP_PR_NEG_4')
    test.cleanup()
    utils_libs.report_results('test_gbp_pr_neg', 'test_results.txt')
    sys.exit(1)


class test_gbp_pr_neg(object):

    # Initialize logging
    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
        level=logging.WARNING)
    _log = logging.getLogger(__name__)
    cmd = 'rm /tmp/test_gbp_pr_neg.log'
    commands.getoutput(cmd)
    hdlr = logging.FileHandler('/tmp/test_gbp_pr_neg.log')
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    _log.addHandler(hdlr)
    _log.setLevel(logging.INFO)
    _log.setLevel(logging.DEBUG)

    def __init__(self):
        """
        Init def
        """
        self._log.info("\n START OF GBP POLICY_RULE NEGATIVE TESTSUITE")
        self.gbpcfg = config_libs.Gbp_Config()
        self.gbpverify = verify_libs.Gbp_Verify()
        self.act_name = 'demo_pa'
        self.cls_name = 'demo_pc'
        self.rule_name = 'demo_pr'
        self._log.info('\n## Step 1: Create a PC needed for PR Testing ##')
        self.cls_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'classifier', self.cls_name)
        if self.cls_uuid == 0:
            self._log.info(
                "\nReqd Classifier Create Failed, hence GBP Policy Rule "
                "Negative Test Suite Run ABORTED\n")
            os._exit(1)
        self._log.info('\n## Step 1: Create a PA needed for PR Testing ##')
        self.act_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'action', self.act_name)
        if self.act_uuid == 0:
            self._log.info(
                "\nReqd Action Create Failed, hence GBP Policy Rule "
                "Negative Test Suite Run ABORTED\n")
            os._exit(1)

    def cleanup(self, tc_name=''):
        if tc_name != '':
            self._log.info('Testcase %s: FAILED' % (tc_name))
        for obj in ['rule', 'classifier', 'action']:
            self.gbpcfg.gbp_del_all_anyobj(obj)

    def test_gbp_pr_neg_1(self):

        self._log.info(
            "\n###################################################\n"
            "TESTCASE_GBP_PR_NEG_1: TO CREATE/VERIFY a POLICY RULE with "
            "INVALID PC\n"
            "TEST_STEP::\n"
            "Create Policy Rule Object with Invalid PC\n"
            "Verify PR creation failed and was rolled back\n"
            "###################################################\n")

        # Testcase work-flow starts
        self._log.info("\n## Step 1: Create Rule with Invalid PC##\n")
        if self.gbpcfg.gbp_policy_cfg_all(
                1, 'rule', self.rule_name, classifier="INVALID") != 0:
            self._log.info(
                "\n## Step 1: Create Policy Rule with Invalid Policy "
                "Classifier did NOT Fail")
            return 0
        self._log.info("\n## Step 1A: Verify Policy Rule has been rolled back")
        if self.gbpverify.gbp_policy_verify_all(
                1, 'rule', self.rule_name) != 0:
            self._log.info(
                "\n## Step 1B: Verify Policy Rule did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PR_NEG_1: PASSED")
        return 1

    def test_gbp_pr_neg_2(self):

        self._log.info(
            "\n#################################################\n"
            "TESTCASE_GBP_PR_NEG_2: TO CREATE/VERIFY/ POLICY RULE with "
            "VALIC PC but INVALID PA\n"
            "TEST_STEP::\n"
            "Create Policy Rule Object with Valid PC but Invalid PA\n"
            "Verify the Policy Rule creation fails and config is rolled back\n"
            "#################################################\n")

        # Testcase work-flow starts
        self._log.info(
            "\n## Step 1: Create Policy Rule with Valid PC & Invalid PA ##")
        if self.gbpcfg.gbp_policy_cfg_all(
                1,
                'rule',
                self.rule_name,
                classifier=self.cls_name,
                action='INVALID') != 0:
            self._log.info(
                "\n## Step 1: Create Policy Rule with Invalid PA did NOT Fail")
            return 0
        self._log.info("\n## Step 1A: Verify Policy Rule has been rolled back")
        if self.gbpverify.gbp_policy_verify_all(
                1, 'rule', self.rule_name) != 0:
            self._log.info(
                "\n## Step 1A: Verify Policy Rule did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PR_NEG_2: PASSED")
        return 1

    def test_gbp_pr_neg_3(self):

        self._log.info(
            "\n################################################\n"
            "TESTCASE_GBP_PR_NEG_3: TO CREATE/UPDATE/VERIFY/ POLICY RULE "
            "with Invalid PC and PA ##\n"
            "TEST_STEP::\n"
            "Create Policy Rule with Valid PC and Valid PR\n"
            "Update the Policy Rule's PA by an Invalid PA\n"
            "Verify the Policy Rule's Update failed and config rolled back "
            "to original attr values\n"
            "Update the Policy Rule's PC by an Invalid PC\n"
            "Verify the Policy Rule's Update failed and config rolled back "
            "to original attr values\n"
            "#################################################\n")

        # Testcase work-flow starts
        self._log.info('\n## Step 1: Create Policy Rule with PA and PC##\n')
        rule_uuid = self.gbpcfg.gbp_policy_cfg_all(
            1, 'rule', self.rule_name, classifier=self.cls_name,
            action=self.act_name)
        if rule_uuid != 0:
            self._log.info(
                "Step 1: Create Rule Passed, UUID == %s\n" %
                (rule_uuid))
        else:
            self._log.info("# Step 1: Create Rule == Failed")
            return 0
        self._log.info(
            '\n## Step 2: Update Policy Rule with Invalid PA and Invalid '
            'PC one at a time ##\n')
        attrib_list = [{'classifier': 'INVALID'}, {'action': 'INVALID'}]
        for attr_val in attrib_list:
            if self.gbpcfg.gbp_policy_cfg_upd_all(
                    'rule', rule_uuid, attr_val) != 0:
                self._log.info(
                    "\nStep 2: Updating Policy Rule's Attribute %s with "
                    "Invalid Value did NOT Fail" %
                    (attr_val))
                return 0
        # Verify starts
        if self.gbpverify.gbp_policy_verify_all(
                1,
                'rule',
                rule_uuid,
                name=self.rule_name,
                policy_classifier_id=self.cls_uuid,
                policy_actions=self.act_uuid) == 0:
            self._log.info(
                "# Step 2B: Verify Policy Rule Updated did NOT roll back")
            return 0
        self._log.info("\n## TESTCASE_GBP_PR_NEG_3: PASSED")
        return 1

    def test_gbp_pr_neg_4(self):

        self._log.info(
            "\n###############################################\n"
            "TESTCASE_GBP_PR_NEG_4: DELETE NON-EXISTENT/INVALID POLICY RULE \n"
            "TEST_STEP::\n"
            "Delete unknown/invalid policy-rule\n"
            "##############################################\n")

        self._log.info("\n## Step 1: Delete non-existent Polic Rule  ##")
        if self.gbpcfg.gbp_policy_cfg_all(0, 'rule', 'INVALID') != 0:
            self._log.info(
                "\n## Step 1: Delete Non-existent policy rule did NOT Fail")
            return 0
        self._log.info("\n## TESTCASE_GBP_PR_NEG_4: PASSED")
        return 1


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""ncp_implementation
"""

# revision identifiers, used by Alembic.
revision = 'd08627f64e37'
down_revision = '9744740aa75c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'ncp_node_to_driver_mapping',
        sa.Column('servicechain_node_id', sa.String(length=36),
                  nullable=False),
        sa.Column('driver_name', sa.String(length=36)),
        sa.Column('servicechain_instance_id', sa.String(length=36)),
        sa.PrimaryKeyConstraint(
            'servicechain_node_id', 'servicechain_instance_id'),
        sa.ForeignKeyConstraint(['servicechain_node_id'],
                                ['sc_nodes.id'], ondelete='CASCADE'),
        sa.ForeignKeyConstraint(['servicechain_instance_id'],
                                ['sc_instances.id'], ondelete='CASCADE')
    )

    op.create_table(
        'ncp_service_targets',
        sa.Column('servicechain_node_id', sa.String(length=36),
                  nullable=False),
        sa.Column('servicechain_instance_id', sa.String(length=36)),
        sa.Column('policy_target_id', sa.String(length=36)),
        sa.Column('relationship', sa.String(length=25)),
        sa.Column('position', sa.Integer),

        sa.PrimaryKeyConstraint(
            'servicechain_node_id', 'servicechain_instance_id',
            'policy_target_id'),
        sa.ForeignKeyConstraint(['policy_target_id'],
                                ['gp_policy_targets.id'], ondelete='CASCADE')
    )


def downgrade():
    pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.contrib.nfp.configurator.drivers.vpn.vyos import (
    vyos_vpn_constants)

""" Implements fake objects for assertion.

"""


class VPNTestData(object):
    '''
    Class which contains the required dictionaries to perform
    vpn ipsec site conn
    '''

    def __init__(self):

        self.context_device = {'notification_data': {},
                               'resource': 'interfaces',
                               'service_vm_context': {'vyos': {
                                             'username': 'username',
                                             'password': 'password'}}}
        self.sc = 'sc'
        self.conf = 'conf'
        self.msg = 'msg'
        self.drivers = 'drivers'
        self.svc = {' ': ' '}
        self.vm_mgmt_ip = '192.168.20.75'
        self.service_vendor = 'vyos'
        self.source_cidrs = '11.0.0.0/24'
        self.destination_cidr = 'destination_cidr'
        self.gateway_ip = '11.0.0.254'
        self.url = 'http://192.168.20.75:8888'
        self.vpn_vpnsvc_error = [{
            'status': 'ERROR',
            'updated_pending_status': True,
            'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'}]
        self.vpn_vpnsvc_active = [{
            'status': 'ACTIVE',
            'updated_pending_status': True,
            'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'}]
        self.ipsec_vpnsvc_status = [{
            'status': 'ACTIVE',
            'ipsec_site_connections': {
                                    'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af6': {
                                        'status': 'INIT',
                                        'updated_pending_status': True}},
            'updated_pending_status': False,
            'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'}]
        self.url_for_add_inte = "%s/add_rule" % self.url
        self.url_for_del_inte = "%s/delete_rule" % self.url
        self.url_for_del_stc_route = "%s/delete-stitching-route" % self.url
        self.url_for_add_src_route = "%s/add-source-route" % self.url
        self.url_for_del_src_route = "%s/delete-source-route" % self.url
        self.url_create_ipsec_conn = "%s/create-ipsec-site-conn" % self.url
        self.url_update_ipsec_conn = "%s/update-ipsec-site-conn" % self.url
        self.url_delete_ipsec_conn = (
            "%s/delete-ipsec-site-conn?peer_address=1.103.2.2" % self.url)
        self.url_create_ipsec_tunnel = "%s/create-ipsec-site-tunnel" % self.url
        self.url_delete_ipsec_tunnel = "%s/delete-ipsec-site-tunnel" % self.url
        self.url_get_ipsec_tunnel = "%s/get-ipsec-site-tunnel-state" % self.url
        self.data_for_interface = {"stitching_mac": "00:0a:95:9d:68:25",
                                   "provider_mac": "00:0a:95:9d:68:16"}
        self.data_for_add_src_route = [{"source_cidr": "1.2.3.4/24",
                                        "gateway_ip": "1.2.3.4/24"},
                                       {"source_cidr": "1.2.3.4/24",
                                        "gateway_ip": "1.2.3.4/24"}]
        self.data_for_del_src_route = [{"source_cidr": "1.2.3.4/24"},
                                       {"source_cidr": "1.2.3.4/24"}]
        self.conn_id = 'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af6'
        self.data_ = {"local_cidr": "11.0.6.0/24",
                      "peer_address": "1.103.2.2",
                      "peer_cidrs": ["141.0.0.1/24"]}
        self.data__ = {"local_cidr": "11.0.6.0/24",
                       "peer_address": "1.103.2.2",
                       "peer_cidr": "141.0.0.0/24"}
        self.timeout = vyos_vpn_constants.REST_TIMEOUT

        self.ipsec_vpn_create = ['fip=192.168.20.75',
                                 'tunnel_local_cidr=11.0.6.0/24',
                                 'user_access_ip=1.103.2.172',
                                 'fixed_ip=192.168.0.3',
                                 'standby_fip=1.103.1.21',
                                 'service_vendor=vyos',
                                 'stitching_cidr=192.168.0.0/28',
                                 'stitching_gateway=192.168.0.1',
                                 'mgmt_gw_ip=30.0.0.254']

        self.ipsec_delete = ['fip=192.168.20.75',
                             'tunnel_local_cidr=11.0.2.0/24',
                             'user_access_ip=1.103.2.178',
                             'fixed_ip=192.168.0.2',
                             'standby_fip=', 'service_vendor=vyos',
                             'stitching_cidr=192.168.0.0/28',
                             'stitching_gateway=192.168.0.1',
                             'mgmt_gw_ip=30.0.0.254']

        self.ipsec_data = {
            'service': {
                'router_id': '73c64bb0-eab9-4f37-85d0-7c8b0c15ed06',
                'status': 'ACTIVE',
                'name': 'VPNService',
                'admin_state_up': True,
                'subnet_id': '7f42e3e2-80a6-4212-9f49-48194ba58fd9',
                'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                'cidr': '30.0.0.0/28',
                'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835',
                'description': (
                        ";".join(self.ipsec_vpn_create))},

            'siteconns': [
                {'connection': {
                    'status': 'INIT',
                    'psk': 'secret',
                    'initiator': 'bi-directional',
                    'access_ip': '1.103.2.172',
                    'name': 'IPsecSiteConnection',
                    'admin_state_up': True,
                    'stitching_fixed_ip': '192.168.0.3',
                    'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                    'description': (
                            ";".join(self.ipsec_vpn_create)),
                    'auth_mode': 'psk',
                    'peer_cidrs': ['141.0.0.0/24'],
                    'mtu': 1500,
                    'ikepolicy_id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
                    'dpd': {'action': 'hold',
                            'interval': 30,
                            'timeout': 120},
                    'route_mode': 'static',
                    'vpnservice_id': (
                        '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'),
                    'peer_address': '1.103.2.2',
                    'peer_id': '192.168.104.228',
                    'id': 'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af6',
                    'tunnel_local_cidr': '11.0.6.0/24',
                    'ipsecpolicy_id': (
                        'b45d99b8-c38b-44ce-9ec8-ba223a83fb46',)},

                 'ipsecpolicy': {
                    'encapsulation_mode': 'tunnel',
                    'encryption_algorithm': '3des',
                    'pfs': 'group5',
                    'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                    'name': 'IPsecPolicy',
                    'transform_protocol': 'esp',
                    'lifetime': {'units': 'seconds', 'value': 3600},
                    'id': 'b45d99b8-c38b-44ce-9ec8-ba223a83fb46',
                    'auth_algorithm': 'sha1',
                    'description': 'My new IPsec policy',
                },

                    'ikepolicy': {
                    'encryption_algorithm': '3des',
                        'pfs': 'group5',
                        'name': 'IKEPolicy',
                        'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                        'lifetime': {'units': 'seconds', 'value': 3600},
                        'description': 'My new IKE policy',
                        'ike_version': 'v1',
                        'id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
                        'auth_algorithm': 'sha1',
                        'phase1_negotiation_mode': 'main',
                }}]}

        self.svc_context = {
            'service': {
                'router_id': '73c64bb0-eab9-4f37-85d0-7c8b0c15ed06',
                'status': 'ACTIVE',
                'name': 'VPNService',
                'admin_state_up': True,
                'subnet_id': '7f42e3e2-80a6-4212-9f49-48194ba58fd9',
                'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                'cidr': '30.0.0.0/28',
                'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835',
                'description': ";".join(self.ipsec_vpn_create),
            },
            'siteconns': [
                {'connection': {
                    'status': 'INIT',
                    'psk': 'secret',
                    'initiator': 'bi-directional',
                    'name': 'IPsecSiteConnection',
                    'admin_state_up': True,
                    'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                    'description': ";".join(self.ipsec_vpn_create),
                    'auth_mode': 'psk',
                    'peer_cidrs': ['141.0.0.0/24'],
                    'mtu': 1500,
                    'ikepolicy_id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
                    'dpd': {'action': 'hold',
                            'interval': 30,
                            'timeout': 120
                            },
                    'route_mode': 'static',
                    'vpnservice_id': (
                        '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'),
                    'peer_address': '1.103.2.2',
                    'peer_id': '192.168.104.228',
                    'id': 'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af6',
                    'ipsecpolicy_id': (
                        'b45d99b8-c38b-44ce-9ec8-ba223a83fb46'),
                },
                    'ipsecpolicy': {
                    'encapsulation_mode': 'tunnel',
                    'encryption_algorithm': '3des',
                    'pfs': 'group5',
                    'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                    'name': 'IPsecPolicy',
                            'transform_protocol': 'esp',
                            'lifetime': {'units': 'seconds', 'value': 3600},
                            'id': 'b45d99b8-c38b-44ce-9ec8-ba223a83fb46',
                            'auth_algorithm': 'sha1',
                            'description': 'My new IPsec policy',
                },
                    'ikepolicy': {
                    'encryption_algorithm': '3des',
                    'pfs': 'group5',
                    'name': 'IKEPolicy',
                            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                            'lifetime': {'units': 'seconds', 'value': 3600},
                            'description': 'My new IKE policy',
                            'ike_version': 'v1',
                            'id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
                            'auth_algorithm': 'sha1',
                            'phase1_negotiation_mode': 'main',
                }}]}

        self.subnet = [{
            'name': 'apic_owned_res_2b0f246b-b0fc-4731-9245-1bd9ac2bd373',
            'enable_dhcp': None,
            'network_id': 'b7432a1c-66a7-45ff-b317-4bbef9449740',
            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
            'dns_nameservers': [],
            'gateway_ip': '192.168.0.1',
            'ipv6_ra_mode': None,
            'allocation_pools': [{'start': '192.168.0.2',
                                  'end': '192.168.0.14'}],
            'host_routes': [],
            'ip_version': 4,
            'ipv6_address_mode': None,
            'cidr': '30.0.0.0/28',
            'id': '7f42e3e2-80a6-4212-9f49-48194ba58fd9',
        }]

        self.vpnservice = [{
            'router_id': '73c64bb0-eab9-4f37-85d0-7c8b0c15ed06',
            'status': 'ACTIVE',
            'name': 'VPNService',
            'admin_state_up': True,
            'subnet_id': '7f42e3e2-80a6-4212-9f49-48194ba58fd9',
            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
            'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835',
            'description': ";".join(self.ipsec_vpn_create),
        }]
        self.ipsec_site_connection = [{
            'status': 'INIT',
            'psk': 'secret',
            'initiator': 'bi-directional',
            'name': 'IPsecSiteConnection',
            'admin_state_up': True,
            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
            'auth_mode': 'psk',
            'peer_cidrs': ['141.0.0.0/24'],
            'mtu': 1500,
            'ikepolicy_id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
            'vpnservice_id': (
                            '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'),
            'dpd': {'action': 'hold',
                    'interval': 30,
                    'timeout': 120},
            'route_mode': 'static',
            'ipsecpolicy_id': (
                'b45d99b8-c38b-44ce-9ec8-ba223a83fb46'),
            'peer_address': '1.103.2.2',
            'peer_id': '192.168.104.228',
            'id': 'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af6',
            'description': ";".join(self.ipsec_vpn_create),
        }]

        self.ipsec_site_connection_delete = [{
            u'status': u'INIT',
            u'psk': u'secret',
            u'initiator': u'bi-directional',
            u'name': u'site_to_site_connection1',
            u'admin_state_up': True,
            u'tenant_id': u'564aeb9ebd694468bfb79a69da887419',
            u'auth_mode': u'psk',
            u'peer_cidrs': [u'11.0.0.0/24'],
            u'mtu': 1500,
            u'ikepolicy_id': (
                u'7a88b9f4-70bf-4184-834d-6814f264d331'),
            u'vpnservice_id': (
                u'3d453be6-7ddc-4812-a4a7-3299f9d3d29e'),
            u'dpd': {u'action': u'hold',
                     u'interval': 30,
                     u'timeout': 120},
            u'route_mode': u'static',
            u'ipsecpolicy_id': (
                u'03839460-1519-46ab-a073-b74314c06ec3'),
            u'peer_address': u'1.103.2.2',
            u'peer_id': u'1.103.2.2',
                        u'id': u'4dae3c91-0d0a-4ba5-9269-d0deab653316',
                        u'description': ";".join(self.ipsec_delete),
        }]

        self.ikepolicies = [{
            'encryption_algorithm': '3des',
            'pfs': 'group5',
            'name': 'IKEPolicy',
            'phase1_negotiation_mode': 'main',
            'lifetime': {'units': 'seconds', 'value': 3600},
            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
            'ike_version': 'v1',
            'id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
            'auth_algorithm': 'sha1',
            'description': 'My new IKE policy',
        }]

        self.ipsecpolicies = [{
            'encapsulation_mode': 'tunnel',
            'encryption_algorithm': '3des',
            'pfs': 'group5',
            'lifetime': {'units': 'seconds', 'value': 3600},
            'name': 'IPsecPolicy',
            'transform_protocol': 'esp',
            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
            'id': 'b45d99b8-c38b-44ce-9ec8-ba223a83fb46',
            'auth_algorithm': 'sha1',
            'description': 'My new IPsec policy',
        }]

        self.context = {
            'domain': None,
            'project_name': None,
            'tenant_name': u'services',
            'project_domain': None,
            'timestamp': '2016-03-03 09:19:05.381231',
            'auth_token': u'0711af29a389492cb799e096a003a760',
            'resource_uuid': None,
            'is_admin': True,
            'user': u'19e278f3c3fa43e3964b057bc73cf7d7',
            'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
            'read_only': False,
            'project_id': 'b',
            'user_id': 'a',
            'show_deleted': False,
            'roles': [u'admin', u'heat_stack_owner'],
            'user_identity': 'a b - - -',
            'tenant_id': u'9f1663d116f74a01991ad66aaa8756c5',
            'request_id': u'req-da8765fb-4eb4-4f4f-9ebb-843ad1d752bd',
            'user_domain': None,
            'user_name': u'neutron',
            'agent_info': {'context': {},
                           'resource': {}},
            "resource_data": {
                    "forward_route": True,
                    "tenant_id": "ac33b4c2d80f485a86ea515c09c74949",
                    "nfs": [{
                        "role": "master",
                        "svc_mgmt_fixed_ip": "192.168.20.75",
                        "networks": [
                            {"cidr": "11.0.2.0/24",
                             "gw_ip": "",
                             "type": "provider",
                             "ports": [{
                                "mac": "fa:16:3e:d9:4c:33",
                                "fixed_ip": "11.0.1.1",
                                "floating_ip": ""}]},
                            {"cidr": "192.168.0.0/28",
                             "gw_ip": "192.168.0.1 ",
                             "type": "stitching",
                             "ports": [{
                                 "mac": "fa:16:3e:da:ca:4d",
                                 "fixed_ip": "192.168.0.3",
                                 "floating_ip": ""}]}
                                        ]}]}}

    def get_header(self):
        header = {'username': 'username',
                  'password': 'password',
                  'Content-Type': 'application/json'}
        return header

    def make_service_context(self, operation_type=None):
        '''
        Prepares a simple service_info dictionary and appends it to
        context dictionary

        '''

        agent_info = {'context': {'service_vm_context': {
                                'vyos': {
                                    'username': 'username',
                                    'password': 'password'}}}}
        self.service_info = {}
        self.service_info.update({'vpnservices': self.vpnservice})
        if operation_type is None:
            self.service_info.update({'ikepolicies': self.ikepolicies})
            self.service_info.update({'ipsecpolicies': self.ipsecpolicies})
            self.service_info.update({'ipsec_site_conns': (
                self.ipsec_site_connection)})

        self.service_info.update({'subnets': self.subnet})
        self.context.update({'service_info': self.service_info})
        self.context.update({'agent_info': agent_info})
        return self.context

    def _create_vpnservice_obj(self):
        '''
        Return the fake dictionary for vpnservice creation
        '''

        return {
            'rsrc_type': 'vpn_service',
            'rsrc_id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835',
            'resource': {
                'router_id': '73c64bb0-eab9-4f37-85d0-7c8b0c15ed06',
                'status': 'ACTIVE',
                'name': 'VPNService',
                'admin_state_up': True,
                'subnet_id': '7f42e3e2-80a6-4212-9f49-48194ba58fd9',
                'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                'id': '36cd27d5-8ad0-4ed7-8bbe-57c488a17835',
                'description': ";".join(self.ipsec_vpn_create),
            },
            'svc_type': 'ipsec',
            'service_vendor': 'vyos',
            'reason': 'create',
        }

    def _create_ipsec_site_conn_obj(self):
        '''
        Return the fake dictionary for ipsec site conn creation
        '''

        return {
            'rsrc_type': 'ipsec_site_connection',
            'rsrc_id': 'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af9',
            'resource': {
                'status': 'INIT',
                'psk': 'secret',
                'initiator': 'bi-directional',
                'name': 'IPsecSiteConnection',
                'admin_state_up': True,
                'tenant_id': '9f1663d116f74a01991ad66aaa8756c5',
                'auth_mode': 'psk',
                'peer_cidrs': ['141.0.0.1/24'],
                'mtu': 1500,
                'ikepolicy_id': '31b79141-3d21-473f-b104-b811bb3ac1fd',
                'vpnservice_id': (
                    '36cd27d5-8ad0-4ed7-8bbe-57c488a17835'),
                'dpd': {'action': 'hold',
                        'interval': 30,
                        'timeout': 120},
                'route_mode': 'static',
                'ipsecpolicy_id': (
                    'b45d99b8-c38b-44ce-9ec8-ba223a83fb46'),
                'peer_address': '1.103.2.2',
                'peer_id': '141.0.0.2',
                'id': 'ac3a0e54-cdf2-4ea7-ac2f-7c0225ab9af9',
                'description': ";".join(self.ipsec_vpn_create),
            },
            'svc_type': 'ipsec',
            'service_vendor': 'vyos',
            'reason': 'create',
        }

    def _delete_ipsec_site_conn_obj(self):
        '''
        Return the fake dictionary for ipsec site conn deletion
        '''

        return {
            u'rsrc_type': u'ipsec_site_connection',
            u'rsrc_id': u'4dae3c91-0d0a-4ba5-9269-d0deab653316',
            u'resource': {
                u'status': u'INIT',
                u'psk': u'secret',
                u'initiator': u'bi-directional',
                u'name': u'site_to_site_connection1',
                u'admin_state_up': True,
                u'tenant_id': u'564aeb9ebd694468bfb79a69da887419',
                u'auth_mode': u'psk',
                u'peer_cidrs': [u'11.0.0.0/24'],
                u'mtu': 1500,
                u'ikepolicy_id': (
                        u'7a88b9f4-70bf-4184-834d-6814f264d331'),
                u'vpnservice_id': (
                    u'3d453be6-7ddc-4812-a4a7-3299f9d3d29e'),
                u'dpd': {u'action': u'hold',
                         u'interval': 30,
                         u'timeout': 120},
                u'route_mode': u'static',
                u'ipsecpolicy_id': (
                    u'03839460-1519-46ab-a073-b74314c06ec3'),
                u'peer_address': u'1.103.2.2',
                u'peer_id': u'1.103.2.2',
                u'id': u'4dae3c91-0d0a-4ba5-9269-d0deab653315',
                u'description': ";".join(self.ipsec_delete),
            },
            u'svc_type': u'ipsec',
            u'service_vendor': u'vyos',
            u'reason': u'delete',
        }

    def _update_ipsec_site_conn_obj(self):
        '''
        Return the fake dictionary for ipsec site conn updation
        '''

        return {
            u'rsrc_type': u'ipsec_site_connection',
            u'rsrc_id': u'4dae3c91-0d0a-4ba5-9269-d0deab653316',
            u'resource': {
                u'status': u'INIT',
                u'psk': u'secret',
                u'initiator': u'bi-directional',
                u'name': u'site_to_site_connection1',
                u'admin_state_up': True,
                u'tenant_id': u'564aeb9ebd694468bfb79a69da887419',
                u'auth_mode': u'psk',
                u'peer_cidrs': [u'11.0.0.0/24'],
                u'mtu': 1500,
                u'ikepolicy_id': (
                        u'7a88b9f4-70bf-4184-834d-6814f264d331'),
                u'vpnservice_id': (
                    u'3d453be6-7ddc-4812-a4a7-3299f9d3d29e'),
                u'dpd': {u'action': u'hold',
                         u'interval': 30,
                         u'timeout': 120},
                u'route_mode': u'static',
                u'ipsecpolicy_id': (
                    u'03839460-1519-46ab-a073-b74314c06ec3'),
                u'peer_address': u'1.103.2.2',
                u'peer_id': u'1.103.2.2',
                u'id': u'4dae3c91-0d0a-4ba5-9269-d0deab653315',
                u'description': ";".join(self.ipsec_vpn_create),
            },
            u'svc_type': u'ipsec',
            u'service_vendor': u'vyos',
            u'reason': u'update',
        }

    def make_resource_data(self, operation=None, service_type=None):
        '''
        Prepares a simple resource_data dictionary of respective service
        '''
        if operation is 'delete':
            return self._delete_ipsec_site_conn_obj()
        if operation is 'update':
            return self._update_ipsec_site_conn_obj()

        if operation == 'create' and service_type == 'ipsec':
            return self._create_ipsec_site_conn_obj()
        else:
            return self._create_vpnservice_obj()

    def fake_resource_data(self):
        '''
        A sample keyword arguments for configurator
        Returns: resource_data
        '''

        resource_data = {
                'forward_route': True,
                'tenant_id': 'ac33b4c2d80f485a86ea515c09c74949',
                'fail_count': 0,
                'nfds': [{
                    'role': 'master',
                    'networks': [{
                        'cidr': '1.2.3.4/24',
                        'gw_ip': '',
                        'type': 'provider',
                        'ports': [{
                            'fixed_ip': '1.2.3.4/24',
                            'mac': '00:0a:95:9d:68:16',
                            'floating_ip': ''
                        }]
                    }, {
                        'cidr': '1.2.3.4/24',
                        'gw_ip': '1.2.3.4/24',
                        'type': 'stitching',
                        'ports': [{
                            'mac': '00:0a:95:9d:68:25',
                            'floating_ip': '',
                            'fixed_ip': '1.2.3.4/24'
                        }]
                    }],
                    'svc_mgmt_fixed_ip': '192.168.20.75',
                    'periodicity': 'initial',
                    'vmid': 'b238e3f12fb64ebcbda2b3330700bf00'
                }]}

        '''resource_data = {'service_type': 'vpn',
                         'vm_mgmt_ip': '192.168.20.75',
                         'mgmt_ip': '192.168.20.75',
                         'source_cidrs': ['1.2.3.4/24'],
                         'destination_cidr': ['1.2.3.4/24'],
                         'gateway_ip': '1.2.3.4/24',
                         'titching_ip': '1.2.3.4/24',
                         'stitching_cidr': '1.2.3.4/24',
                         'provider_ip': '1.2.3.4/24',
                         'provider_ip': '1.2.3.4/24',
                         'provider_interface_position': '1',
                         'stitching_interface_position': '3',
                         'request_info': 'some_id',
                         'periodicity': 'initial',
                         'provider_mac': '00:0a:95:9d:68:16',
                         'stitching_mac': '00:0a:95:9d:68:25',
                         'context': {'notification_data': 'hello'}
                         }'''
        return resource_data


class FakeEvent(object):
    '''
    Implements a fake event class for process framework to use to create
    the fake event object.
    '''

    def __init__(self):
        self.dict_obj = VPNTestData()
        self.data = {
            'context': self.dict_obj.make_service_context(),
            'resource_data': self.dict_obj._create_ipsec_site_conn_obj()
        }
